from datetime import datetime
from typing import Any, Callable, List, Optional, Union

import dask
import xarray
from xarray import DataArray, Dataset
from xclim.core import calendar

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.cf_calendar import CfCalendar
from icclim.models.frequency import Frequency, SliceMode
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation


class CfVariable:
    """
    CfVariable groups together two xarray DataArray for the same variable.
    One represent the whole studied period. The other is only the in base period used by
    percentile based indices to compute percentiles.

    Parameters
    ----------
    da: DataArray
        The variable studied.
    in_base_da: DataArray
        The variable studied limited to the in base period.
    """

    da: DataArray
    in_base_da: DataArray

    def __init__(self, da: DataArray, in_base_da: DataArray) -> None:
        self.da = da
        self.in_base_da = in_base_da


class IndexConfig:
    """
    Configuration class for ECA&D indices.

    Parameters
    ----------
    freq: Frequency
        The expected resampling frequency of the output.
    cf_variables: List[CfVariable]
        List of CfVariable necessary to compute the index.
    save_percentile: bool = False
        On percentile based indices, if True, this saves the percentile in the output
        netcdf.
    is_percent: bool = False
        On indices resulting in a numbers of days, if True, this converts the results to
        % of the sampling frequency
    netcdf_version: NetcdfVersion
        Netcdf version to be used when creating the output
    window: Optional[int]
        On indices relying on a rolling window of days, configure the window width.
    threshold: Optional[float]
        On indices relying on a threshold, configure the threshold value. Unit less.
        The unit "degC" is added by icclim.
    transfer_limit_Mbytes: Optional[int]
        The dask maximum chunk size.
    out_unit: Optional[str]
        The output unit overriding Xclim results.
    callback: Optional[Callable]
        A callable to produce a progress bar
    """

    freq: Frequency
    cf_variables: List[CfVariable]
    save_percentile: bool = False
    is_percent: bool = False
    netcdf_version: NetcdfVersion
    window: Optional[int]
    threshold: Optional[float]
    transfer_limit_Mbytes: Optional[int]
    out_unit: Optional[str]
    callback: Optional[Callable]

    def __init__(
        self,
        ds: Dataset,
        slice_mode: SliceMode,
        var_name: List[str],
        netcdf_version: Union[str, NetcdfVersion],
        index: Optional[Any],  # EcadIndex proper typing causes circular dependency
        save_percentile: bool = False,
        only_leap_years: bool = False,
        ignore_Feb29th: bool = False,
        window_width: Optional[int] = 5,
        time_range: Optional[List[datetime]] = None,
        base_period_time_range: Optional[List[datetime]] = None,
        transfer_limit_Mbytes: Optional[int] = None,
        threshold: Optional[float] = None,
        out_unit: Optional[str] = None,
        interpolation: Optional[
            QuantileInterpolation
        ] = QuantileInterpolation.MEDIAN_UNBIASED,
        callback: Optional[Callable] = None,
    ):
        self.freq = Frequency.lookup(slice_mode)
        if time_range is not None:
            time_range = [x.strftime("%Y-%m-%d") for x in time_range]
        if base_period_time_range is not None:
            base_period_time_range = [
                x.strftime("%Y-%m-%d") for x in base_period_time_range
            ]
        self.cf_variables = [
            _build_cf_variable(
                da=ds[cf_var_name],
                time_range=time_range,
                ignore_Feb29th=ignore_Feb29th,
                base_period_time_range=base_period_time_range,
                only_leap_years=only_leap_years,
                transfer_limit_Mbytes=transfer_limit_Mbytes,
                index=index,
            )
            for cf_var_name in var_name
        ]
        self.window = window_width
        self.save_percentile = save_percentile
        self.is_percent = out_unit == "%"
        self.out_unit = out_unit
        self.transfer_limit_Mbytes = transfer_limit_Mbytes
        if isinstance(netcdf_version, str):
            self.netcdf_version = NetcdfVersion.lookup(netcdf_version)
        else:
            self.netcdf_version = netcdf_version
        self.interpolation = interpolation
        self.threshold = threshold
        self.callback = callback
        self.index = index


def _build_cf_variable(
    da: DataArray,
    time_range: Optional[List[datetime]],
    ignore_Feb29th: bool,
    base_period_time_range: Optional[List[datetime]],
    only_leap_years: bool,
    transfer_limit_Mbytes: Optional[int],
    index: Optional[Any],  # EcadIndex
) -> CfVariable:
    if transfer_limit_Mbytes is not None:
        da = _chunk_data(transfer_limit_Mbytes, da, index)
    out_of_base_da = _build_data_array(da, time_range, ignore_Feb29th)
    if base_period_time_range is not None:
        in_base_da = _build_in_base_da(da, base_period_time_range, only_leap_years)
    else:
        in_base_da = out_of_base_da
    return CfVariable(out_of_base_da, in_base_da)


def _build_data_array(
    original_da: DataArray, time_range: Optional[List[datetime]], ignore_Feb29th: bool
) -> DataArray:
    if time_range is not None:
        if len(time_range) != 2:
            raise InvalidIcclimArgumentError(
                f"The given `time_range` {time_range}"
                f" has {len(time_range)} elements."
                f" It must have exactly 2 dates."
            )
        da = original_da.sel(time=slice(time_range[0], time_range[1]))
        if len(da.time) == 0:
            raise InvalidIcclimArgumentError(
                f"The given `time_range` {time_range} "
                f"is out of the dataset time period: "
                f"{original_da.time.min().dt.floor('D').values} "
                f"- {original_da.time.max().dt.floor('D').values}."
            )
    else:
        da = original_da
    if ignore_Feb29th:
        da = calendar.convert_calendar(da, CfCalendar.NO_LEAP.get_name())  # type:ignore
    return da


def _build_in_base_da(
    original_da: DataArray,
    base_period_time_range: List[datetime],
    only_leap_years: bool,
) -> DataArray:
    # TODO merge with _build_data_array ?
    if len(base_period_time_range) != 2:
        raise InvalidIcclimArgumentError(
            f"The given `base_period_time_range` {base_period_time_range}"
            f" has {len(base_period_time_range)} elements."
            f" It must have exactly 2 dates."
        )
    da = original_da.sel(
        time=slice(base_period_time_range[0], base_period_time_range[1])
    )
    if len(da.time) == 0:
        raise InvalidIcclimArgumentError(
            f"The given `base_period_time_range` {base_period_time_range}"
            f" is out of the sample time bounds: "
            f"{original_da.time.min().dt.floor('D').values} "
            f"- {original_da.time.max().dt.floor('D').values}."
        )
    if only_leap_years:
        da = _reduce_only_leap_years(original_da)
    return da


def _chunk_data(
    transfer_limit_Mbytes: int, da: DataArray, index: Optional[Any]  # EcadIndex
) -> DataArray:
    with dask.config.set({"array.chunk-size": f"{transfer_limit_Mbytes} MB"}):
        chunks = {d: "auto" for d in da.dims}
        return da.chunk(chunks=chunks)


def _reduce_only_leap_years(da: DataArray) -> DataArray:
    reduced_list: List[DataArray] = []
    for _, val in da.groupby(da.time.dt.year):
        if val.time.dt.dayofyear.max() == 366:
            reduced_list.append(val)
    if not reduced_list:
        raise InvalidIcclimArgumentError(
            "No leap year in current dataset. Do not use only_leap_years parameter."
        )
    return xarray.concat(reduced_list, "time")



#! /usr/bin/env python
import ast
import configparser
import logging
import os
from argparse import ArgumentParser

from auxiliary import (
    ConfigSectionMap,
    get_cfg,
    get_input_file_path,
    get_output_file_path,
    get_test_md5hash,
    get_time_ranges,
    get_varnames_from_filenames,
    try_literal_interpretation,
)

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")


# Check command arguments.
description = """Icclim-test wrapper.py
Reads, parses and and executes test cases defined as Python config files"""

parser = ArgumentParser(description=description)
parser.add_argument(
    "-t",
    "--test-config",
    dest="testfiles",
    type=str,
    nargs="+",
    help="Test file or directory with the test cases in Python config format",
)
parser.add_argument(
    "-i",
    "--input",
    dest="input_test_data_dir",
    type=str,
    nargs="?",
    default="None",
    required=False,
    help="Folder for input test data files",
)
parser.add_argument(
    "-o",
    "--output",
    dest="test_output_dir",
    type=str,
    nargs="?",
    default="None",
    required=False,
    help="Output folder for test cases",
)
parser.add_argument(
    "-l",
    "--list-tests",
    action="store_true",
    dest="list_only",
    default=False,
    required=False,
    help="Only list tests, do not execute them",
)
args = parser.parse_args()

for files in args.testfiles:
    if os.path.isfile(files):
        files = [files]
    # Expand files if test dir given
    elif os.path.isdir(files):
        files = [os.path.join(files, file) for file in sorted(os.listdir(files))]
    else:
        raise ValueError(f"files is unexpected value: {files}")

    for file in files:
        config = configparser.ConfigParser()
        config.read(file)
        sections = [sec for sec in config.sections() if sec != "input_output"]

        # Use config file settings for input if not given on cmd-line
        input_test_data_dir = get_cfg(
            args, config, req_section="input_output", req_option="input_test_data_dir"
        )

        test_output_dir = get_cfg(
            args, config, req_section="input_output", req_option="test_output_dir"
        )

        # Create output folder
        if not os.path.exists(args.test_output_dir):
            os.mkdir(args.test_output_dir)

        # Loop sections in test config file
        for section in sections:
            ind = section.split("_")[0]
            logging.info("======> " + section)

            # Process some of the entries from the test config file
            section_dict = ConfigSectionMap(config, section)
            # section_dict = get_callback(section_dict)
            section_dict = get_input_file_path(section_dict, input_test_data_dir)
            section_dict = get_varnames_from_filenames(section_dict)
            section_dict = get_time_ranges(section_dict)
            section_dict = try_literal_interpretation(section_dict, "slice_mode")
            section_dict = try_literal_interpretation(section_dict, "threshold")

            if "user_indices" in section_dict.keys():
                section_dict["user_indices"] = ast.literal_eval(
                    section_dict["user_indices"]
                )

            testid = get_test_md5hash(section_dict, section, test_output_dir)
            section_dict = get_output_file_path(test_output_dir, section, section_dict)

            logging.info("Test ID: " + testid)

            # Run test
            if not args.list_only:
                import icclim.icclim

                icclim.icclim.index(**section_dict)
            logging.info("<====== " + section + "\n")



from typing import Sequence

import numpy as np
import xarray
from xarray import DataArray
from xclim.indices.run_length import rle_1d


def get_longest_run_start_index(
    arr: DataArray,
    window: int = 1,
    dim: str = "time",
) -> DataArray:
    return xarray.apply_ufunc(
        get_index_of_longest_run,
        arr,
        input_core_dims=[[dim]],
        kwargs={"window": window},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )


def get_index_of_longest_run(arr: Sequence[bool], window: int = 1) -> int:
    values, rl, pos = rle_1d(arr)
    if not np.any(values) or np.all(values * rl < window):  # type:ignore
        return 0
    index_of_max = np.nanargmax(
        np.where(values * rl >= window, rl, np.NaN)  # type:ignore
    )
    return pos[index_of_max]  # type:ignore


def get_first_occurrence_index(da: DataArray) -> DataArray:
    """
    Return the index of the first True value in the 3D booleans array along
    time dimension.
    """
    stacked = da.stack(latlon=("lat", "lon"))
    res = stacked.argmax("time")
    return res.unstack()



from enum import Enum
from typing import Any, Callable, Union

from xarray.core.dataarray import DataArray

from icclim.icclim_exceptions import InvalidIcclimArgumentError, MissingIcclimInputError
from icclim.models.user_index_config import UserIndexConfig
from icclim.user_indices import operators


def compute_user_index(config: UserIndexConfig) -> DataArray:
    operation = CalcOperation.lookup(config)
    return operation.compute_fun(config)


def anomaly(config: UserIndexConfig):
    if config.da_ref is None:
        raise MissingIcclimInputError(
            f"You must provide a in base to compute {CalcOperation.ANOMALY.value}."
        )
    return operators.anomaly(
        da=config.cf_vars[0].da,
        da_ref=config.da_ref,
        percent=config.is_percent,
    )


def run_sum(config: UserIndexConfig):
    if config.extreme_mode is None or config.window_width is None:
        raise MissingIcclimInputError(
            "Please provide a extreme mode and a window width."
        )
    return operators.run_sum(
        da=config.cf_vars[0].da,
        extreme_mode=config.extreme_mode,
        window_width=config.window_width,
        coef=config.coef,
        freq=config.freq.panda_freq,
        date_event=config.date_event,
    )


def run_mean(config: UserIndexConfig):
    if config.extreme_mode is None or config.window_width is None:
        raise MissingIcclimInputError(
            "Please provide a extreme mode and a window width."
        )
    return operators.run_mean(
        da=config.cf_vars[0].da,
        extreme_mode=config.extreme_mode,
        window_width=config.window_width,
        coef=config.coef,
        freq=config.freq.panda_freq,
        date_event=config.date_event,
    )


def max_consecutive_event_count(config: UserIndexConfig):
    if config.logical_operation is None or config.thresh is None:
        raise MissingIcclimInputError(
            "Please provide a threshold and a logical operation."
        )
    if isinstance(config.thresh, list):
        raise InvalidIcclimArgumentError(
            f"{CalcOperation.MAX_NUMBER_OF_CONSECUTIVE_EVENTS.value} "
            f"does not support threshold list. Please provide a single threshold."
        )
    return operators.max_consecutive_event_count(
        da=config.cf_vars[0].da,
        in_base_da=config.cf_vars[0].in_base_da,
        logical_operation=config.logical_operation,
        threshold=config.thresh,
        coef=config.coef,
        freq=config.freq.panda_freq,
        date_event=config.date_event,
    )


def count_events(config: UserIndexConfig):
    if config.nb_event_config is None:
        raise MissingIcclimInputError(
            f"{CalcOperation.EVENT_COUNT.value} not properly configure."
            f" Please provide a threshold and a logical operation."
        )
    return operators.count_events(
        das=list(map(lambda x: x.da, config.cf_vars)),
        in_base_das=list(map(lambda x: x.in_base_da, config.cf_vars)),
        logical_operation=config.nb_event_config.logical_operation,
        link_logical_operations=config.nb_event_config.link_logical_operations,
        thresholds=config.nb_event_config.thresholds,
        coef=config.coef,
        freq=config.freq.panda_freq,
        date_event=config.date_event,
    )


def sum(config: UserIndexConfig):
    return operators.sum(
        da=_check_and_get_da(config),
        in_base_da=_check_and_get_in_base_da(config),
        coef=config.coef,
        logical_operation=config.logical_operation,
        threshold=_check_and_get_simple_threshold(config.thresh),
        freq=config.freq.panda_freq,
    )


def mean(config: UserIndexConfig):
    return operators.mean(
        da=_check_and_get_da(config),
        in_base_da=_check_and_get_in_base_da(config),
        coef=config.coef,
        logical_operation=config.logical_operation,
        threshold=_check_and_get_simple_threshold(config.thresh),
        freq=config.freq.panda_freq,
    )


def min(config: UserIndexConfig):
    return operators.min(
        da=_check_and_get_da(config),
        in_base_da=_check_and_get_in_base_da(config),
        coef=config.coef,
        logical_operation=config.logical_operation,
        threshold=_check_and_get_simple_threshold(config.thresh),
        freq=config.freq.panda_freq,
        date_event=config.date_event,
    )


def max(config: UserIndexConfig):
    return operators.max(
        da=_check_and_get_da(config),
        in_base_da=_check_and_get_in_base_da(config),
        coef=config.coef,
        logical_operation=config.logical_operation,
        threshold=_check_and_get_simple_threshold(config.thresh),
        freq=config.freq.panda_freq,
        date_event=config.date_event,
    )


def _check_and_get_simple_threshold(thresh: Any) -> Union[None, str, float, int]:
    if (
        thresh is None
        or isinstance(thresh, str)
        or isinstance(thresh, float)
        or isinstance(thresh, int)
    ):
        return thresh
    else:
        raise InvalidIcclimArgumentError(
            "threshold type must be either None, "
            "a string (for percentile) or a number."
        )


def _check_and_get_da(config: UserIndexConfig) -> DataArray:
    if len(config.cf_vars) == 1:
        return config.cf_vars[0].da
    else:
        raise InvalidIcclimArgumentError(
            f"There must be exactly one variable for {config.calc_operation}."
        )


def _check_and_get_in_base_da(config: UserIndexConfig) -> Union[DataArray, None]:
    if len(config.cf_vars) == 1:
        return config.cf_vars[0].in_base_da
    else:
        raise InvalidIcclimArgumentError(
            f"There must be exactly one variable for {config.calc_operation}"
        )


class CalcOperation(Enum):
    MAX = ("max", max)
    MIN = ("min", min)
    SUM = ("sum", sum)
    MEAN = ("mean", mean)
    EVENT_COUNT = ("nb_events", count_events)
    MAX_NUMBER_OF_CONSECUTIVE_EVENTS = (
        "max_nb_consecutive_events",
        max_consecutive_event_count,
    )
    RUN_MEAN = ("run_mean", run_mean)
    RUN_SUM = ("run_sum", run_sum)
    ANOMALY = ("anomaly", anomaly)

    def __init__(
        self, input_name: str, compute_fun: Callable[[UserIndexConfig], DataArray]
    ):
        self.input_name = input_name
        self.compute_fun = compute_fun

    @staticmethod
    def lookup(config: UserIndexConfig):
        if isinstance(config.calc_operation, CalcOperation):
            return config.calc_operation
        for calc_op in CalcOperation:
            if calc_op.input_name.upper() == config.calc_operation.upper():
                return calc_op
        raise InvalidIcclimArgumentError(
            f"The calc_operation {config.calc_operation} is unknown."
        )



# keep line below to alias "main" module to icclim
from . import main as icclim

# keep line below to expose "main" content in icclim package namespace
from .main import index, indice

__version__ = "5.0.0rc2"



# fmt: off
# flake8: noqa

# ECA&D indices groups
TEMPERATURE_GROUP = "temperature"
HEAT_GROUP = "heat"
COLD_GROUP = "cold"
DROUGHT_GROUP = "drought"
RAIN_GROUP = "rain"
SNOW_GROUP = "snow"
COMPOUND_GROUP = "compound"

# placeholder for user_indices
PERCENTILE_THRESHOLD_STAMP = "p"
WET_DAY_THRESHOLD = 1  # 1mm
PRECIPITATION = "p"
TEMPERATURE = "t"

# percentiles dimension from percentile_doy
PERCENTILES_COORD = "percentiles"
# attribut holding the in_base time bounds
IN_BASE_IDENTIFIER = "reference_epoch"

# Aliases of input variables names. Source: clix-meta
PR = ["pr", "pradjust", "prec", "rr", "precip", "PREC", "Prec", "RR", "PRECIP", "Precip"]
TAS = ["tas", "tavg", "ta", "tasadjust", "tmean", "tm", "tg", "meant", "TMEAN", "Tmean", "TM", "TG", "MEANT", "meanT", "tasmidpoint"]
TAS_MAX = ["tasmax", "tasmaxadjust", "tmax", "tx", "maxt", "TMAX", "Tmax", "TX", "MAXT", "maxT"]
TAS_MIN = ["tasmin", "tasminadjust", "tmin", "tn", "mint", "TMIN", "Tmin", "TN", "MINT", "minT"]



from functools import reduce
from typing import Any, Callable, List, Optional, Union
from warnings import warn

import dask.array
import numpy as np
import xarray
from xarray.core.dataarray import DataArray
from xclim.core.bootstrapping import percentile_bootstrap
from xclim.core.calendar import percentile_doy, resample_doy
from xclim.core.units import convert_units_to, to_agg_units
from xclim.indices.run_length import longest_run

from icclim.icclim_exceptions import (
    InvalidIcclimArgumentError,
    InvalidIcclimOutputError,
)
from icclim.models.constants import (
    PERCENTILE_THRESHOLD_STAMP,
    PERCENTILES_COORD,
    PRECIPITATION,
    TEMPERATURE,
    WET_DAY_THRESHOLD,
)
from icclim.models.user_index_config import (
    ExtremeMode,
    LinkLogicalOperation,
    LogicalOperation,
)
from icclim.user_indices.stat import (
    get_first_occurrence_index,
    get_longest_run_start_index,
)

__all__ = [
    "max",
    "min",
    "sum",
    "mean",
    "count_events",
    "max_consecutive_event_count",
    "run_mean",
    "run_sum",
    "anomaly",
]


def max(
    da: DataArray,
    in_base_da: Optional[DataArray] = None,
    coef: Optional[float] = None,
    logical_operation: Optional[LogicalOperation] = None,
    threshold: Optional[Union[str, float, int]] = None,
    freq: str = "MS",
    date_event: bool = False,
    var_type: Optional[str] = None,
    save_percentile=False,
) -> DataArray:
    result = _apply_coef(coef, da)
    result = _filter_by_threshold(
        result,
        in_base_da,
        logical_operation,
        threshold,
        freq,
        var_type,
        save_percentile,
    )
    resampled = result.resample(time=freq)
    if date_event:
        return _reduce_with_date_event(
            resampled, lambda x: x.argmax("time")  # type:ignore
        )
    else:
        return resampled.max(dim="time")


def min(
    da: DataArray,
    in_base_da: Optional[DataArray] = None,
    coef: float = None,
    logical_operation: LogicalOperation = None,
    threshold: Optional[Union[str, float, int]] = None,
    freq: str = "MS",
    date_event: bool = False,
    var_type: str = None,
    save_percentile=False,
) -> DataArray:
    result = _apply_coef(coef, da)
    result = _filter_by_threshold(
        result,
        in_base_da,
        logical_operation,
        threshold,
        freq,
        var_type,
        save_percentile,
    )
    resampled = result.resample(time=freq)
    if date_event:
        return _reduce_with_date_event(
            resampled, lambda x: x.argmin("time")  # type:ignore
        )
    else:
        return resampled.min(dim="time")


def sum(
    da: DataArray,
    in_base_da: Optional[DataArray] = None,
    coef: float = None,
    logical_operation: LogicalOperation = None,
    threshold: Optional[Union[str, float, int]] = None,
    var_type: str = None,
    freq: str = "MS",
    save_percentile=False,
) -> DataArray:
    result = _apply_coef(coef, da)
    result = _filter_by_threshold(
        result,
        in_base_da,
        logical_operation,
        threshold,
        freq,
        var_type,
        save_percentile,
    )
    return result.resample(time=freq).sum(dim="time")


def mean(
    da: DataArray,
    in_base_da: Optional[DataArray] = None,
    coef: float = None,
    logical_operation: LogicalOperation = None,
    threshold: Optional[Union[str, float, int]] = None,
    var_type: str = None,
    freq: str = "MS",
    save_percentile=False,
) -> DataArray:
    result = _apply_coef(coef, da)
    result = _filter_by_threshold(
        result,
        in_base_da,
        logical_operation,
        threshold,
        freq,
        var_type,
        save_percentile,
    )
    return result.resample(time=freq).mean(dim="time")


def count_events(
    logical_operation: List[LogicalOperation],
    thresholds: List[Union[float, str]],
    das: List[DataArray],
    in_base_das: List[Optional[DataArray]],
    link_logical_operations: LinkLogicalOperation = None,
    coef: float = None,
    var_type: str = None,
    freq: str = "MS",
    date_event: bool = False,
    save_percentile: bool = False,
) -> DataArray:
    percentiles = []
    for i, threshold in enumerate(thresholds):
        if isinstance(threshold, str) and len(in_base_das) > 0:
            in_base_da = in_base_das[i]
            if in_base_da is not None:
                percentiles.append(_get_percentiles(threshold, var_type, in_base_da))
    acc = []
    for i, da in enumerate(das):
        result = _apply_coef(coef, da)
        if len(percentiles) > 0:
            result = _threshold_compare_on_percentiles(
                da=result,
                percentiles=percentiles[i],
                logical_operation=logical_operation[i],
                freq=freq,
                bootstrap=_is_bootstrappable(var_type),
            )
            if save_percentile:
                result.coords[f"percentile_{thresholds[i]}"] = resample_doy(
                    percentiles[i], result
                )
        else:
            result = logical_operation[i].compute(result, thresholds[i])  # type:ignore
        acc.append(result)
    if len(acc) == 1:
        result = acc[0]
    elif link_logical_operations == LinkLogicalOperation.AND_STAMP:
        result = reduce(np.logical_and, acc, True)  # type:ignore
    elif link_logical_operations == LinkLogicalOperation.OR_STAMP:
        result = reduce(np.logical_or, acc, False)  # type:ignore
    else:
        raise NotImplementedError()
    resampled = result.resample(time=freq)
    if date_event:
        return _get_count_events_date_event(resampled)
    return resampled.sum(dim="time")


def max_consecutive_event_count(
    da: DataArray,
    logical_operation: LogicalOperation,
    in_base_da: Optional[DataArray] = None,
    threshold: Optional[Union[str, float, int]] = None,
    coef: float = None,
    freq: str = "MS",
    date_event: bool = False,
    var_type: Optional[str] = None,
    save_percentile=False,
) -> DataArray:
    result = _apply_coef(coef, da)
    if in_base_da is not None and isinstance(threshold, str):
        per = _get_percentiles(threshold, var_type, in_base_da)
        result = _threshold_compare_on_percentiles(
            da=da,
            percentiles=per,
            logical_operation=logical_operation,
            freq=freq,
            bootstrap=_is_bootstrappable(var_type),
        )
        if save_percentile:
            result.coords[PERCENTILES_COORD] = resample_doy(per, result)
    elif isinstance(threshold, float) or isinstance(threshold, int):
        result = logical_operation.compute(da, threshold)
    resampled = result.resample(time=freq)
    if not date_event:
        return resampled.map(longest_run, dim="time")
    acc: List[DataArray] = []
    for label, value in resampled:
        run_length = longest_run(value, dim="time")
        index = get_longest_run_start_index(value, dim="time")
        start = value[index.astype(int)].time
        time_shift = run_length * np.timedelta64(1, "D")
        end = start + time_shift
        coords = dict(
            time=label,
            lat=value.lat,
            lon=value.lon,
            event_date_start=start,
            event_date_end=end,
        )
        acc.append(DataArray(data=run_length, dims=["lat", "lon"], coords=coords))
    result = xarray.concat(acc, "time")
    return to_agg_units(result, da, "count")


def run_mean(
    da: DataArray,
    extreme_mode: ExtremeMode,
    window_width: int,
    coef: float = None,
    freq: str = "MS",
    date_event: bool = False,
) -> DataArray:
    return _run_aggregator(
        da=da,
        extreme_mode=extreme_mode,
        window_width=window_width,
        coef=coef,
        freq=freq,
        date_event=date_event,
        aggregator=lambda da: da.mean(),
    )


def run_sum(
    da: DataArray,
    extreme_mode: ExtremeMode,
    window_width: int,
    coef: float = None,
    freq: str = "MS",
    date_event: bool = False,
) -> DataArray:
    return _run_aggregator(
        da=da,
        extreme_mode=extreme_mode,
        window_width=window_width,
        coef=coef,
        freq=freq,
        date_event=date_event,
        aggregator=lambda da: da.sum(),
    )


def anomaly(da_ref: DataArray, da: DataArray, percent: bool) -> DataArray:
    ref_mean = da_ref.mean(dim="time")
    result: DataArray = da.mean(dim="time") - ref_mean
    result._copy_attrs_from(da_ref)
    if percent:
        result = result / ref_mean * 100
        result.attrs["units"] = "%"
    return result


def _apply_coef(coef: Optional[float], da: DataArray) -> DataArray:
    if coef is not None:
        return da * coef
    return da


def _filter_by_threshold(
    da: DataArray,
    in_base_da: Optional[DataArray],
    logical_operation: Optional[LogicalOperation],
    threshold: Optional[Union[str, float, int]],
    freq: str,
    var_type: Optional[str],
    save_percentile: bool,
) -> DataArray:
    if threshold is None and logical_operation is None:
        return da
    if isinstance(threshold, str):
        if in_base_da is None:
            raise NotImplementedError(
                "When threshold type is str for percentiles, a in_base must be provided"
            )
        per = _get_percentiles(threshold, var_type, in_base_da)
        result = _filter_by_logical_op_on_percentile(
            da=da,
            percentiles=_get_percentiles(threshold, var_type, in_base_da),
            logical_operation=logical_operation,
            freq=freq,
            bootstrap=_is_bootstrappable(var_type),
        )
        if save_percentile:
            result.coords[PERCENTILES_COORD] = resample_doy(per, result)
    elif (
        isinstance(threshold, float) or isinstance(threshold, int)
    ) and logical_operation is not None:
        result = da.where(logical_operation.compute(da, threshold))
    else:
        raise NotImplementedError(
            "threshold type must be on of [str, int, float] and logical_operation must "
            "a LogicalOperation instance"
        )
    if len(result) == 0:
        raise InvalidIcclimOutputError(
            f"The dataset has been emptied by filtering with "
            f"{logical_operation.operator}{threshold}."
        )
    return result


@percentile_bootstrap
def _filter_by_logical_op_on_percentile(
    da: DataArray,
    percentiles: Optional[DataArray],
    logical_operation: Optional[LogicalOperation],
    freq: str = "MS",  # used by percentile_bootstrap
    bootstrap: bool = False,  # used by percentile_bootstrap
) -> DataArray:
    if logical_operation is not None and percentiles is not None:
        percentiles = resample_doy(percentiles, da)
        mask = logical_operation.compute(da, percentiles)
        result = da.where(mask, drop=True)
        if bootstrap:
            result = da.expand_dims(_bootstrap=result._bootstrap)
        return result
    return da


@percentile_bootstrap
def _threshold_compare_on_percentiles(
    da: DataArray,
    percentiles: DataArray,
    logical_operation: LogicalOperation,
    freq: str = "MS",  # used by percentile_bootstrap
    bootstrap: bool = False,  # used by percentile_bootstrap
) -> DataArray:
    percentiles = resample_doy(percentiles, da)
    return logical_operation.compute(da, percentiles)


def _get_percentiles(
    thresh: str, var_type: Optional[str], in_base_da: DataArray
) -> DataArray:
    if thresh.find(PERCENTILE_THRESHOLD_STAMP) == -1:
        raise InvalidIcclimArgumentError(
            "Percentile threshold not properly formatted."
            " Use p as a prefix or suffix of the value for example 90p or p90."
            " For non percentile threshold use a float instead of a string"
        )
    per = float(thresh.replace(PERCENTILE_THRESHOLD_STAMP, ""))
    da_per = in_base_da
    if var_type == PRECIPITATION:
        da_per = convert_units_to(in_base_da, "mm/d")
        da_per = da_per.where(da_per > WET_DAY_THRESHOLD, drop=True)
    percentiles = percentile_doy(arr=da_per, per=per).sel(percentiles=per)
    return percentiles


def _run_aggregator(
    da: DataArray,
    extreme_mode: ExtremeMode,
    window_width: int,
    aggregator: Callable[[Any], DataArray],  # Any should be DataArrayRolling
    coef: float = None,
    freq: str = "MS",
    date_event: bool = False,
) -> DataArray:
    result = _apply_coef(coef, da)
    result = result.rolling(time=window_width)
    resampled = aggregator(result).resample(time=freq)
    if extreme_mode == ExtremeMode.MIN:
        if date_event:
            return _reduce_with_date_event(
                resampled,
                lambda x: x.argmin("time"),  # type:ignore
                window=window_width,
            )
        else:
            return resampled.min(dim="time")
    elif extreme_mode == ExtremeMode.MAX:
        if date_event:
            return _reduce_with_date_event(
                resampled,
                lambda x: x.argmax("time"),  # type:ignore
                window=window_width,
            )
        else:
            return resampled.max(dim="time")
    else:
        raise NotImplementedError()


def _reduce_with_date_event(
    resampled: DataArray,
    reducer: Callable[[DataArray], DataArray],
    window: Optional[int] = None,
) -> DataArray:
    acc: List[DataArray] = []
    for label, value in resampled:
        reduced_result = value.isel(time=reducer(value))
        if window is not None:
            coords = dict(
                time=label,
                lat=value.lat,
                lon=value.lon,
                event_date_start=reduced_result.time,
                event_date_end=reduced_result.time + np.timedelta64(window, "D"),
            )
        else:
            coords = dict(
                time=label,
                lat=value.lat,
                lon=value.lon,
                event_date=reduced_result.time,
            )
        acc.append(DataArray(data=reduced_result, dims=["lat", "lon"], coords=coords))
    return xarray.concat(acc, "time")


def _get_count_events_date_event(resampled):
    if isinstance(resampled, dask.array.Array):
        warn("Computing event_date_start/end when using Dask arrays can be slow.")
    acc: List[DataArray] = []
    for label, sample in resampled:
        # Fixme probably not safe to compute on huge dataset,
        #  it should be fixed with
        #  https://github.com/pydata/xarray/issues/2511
        sample = sample.compute()
        first = sample.isel(time=get_first_occurrence_index(sample)).time
        value_reversed_time = sample[::-1, :, :]
        last = sample.isel(time=get_first_occurrence_index(value_reversed_time)).time
        acc.append(
            DataArray(
                data=sample.sum(dim="time"),
                dims=["lat", "lon"],
                coords=dict(
                    time=label,
                    lat=sample.lat,
                    lon=sample.lon,
                    event_date_start=first,
                    event_date_end=last,
                ),
            )
        )
    return xarray.concat(acc, "time")


def _is_bootstrappable(var_type):
    return var_type == TEMPERATURE



import ast
import copy
import datetime
import hashlib
import os
import re
import sys
import types
from configparser import NoOptionError, NoSectionError

sys.path.append(os.getcwd())


def ConfigSectionMap(Config, section):
    dict1 = {}
    options = Config.options(section)
    for option in options:
        try:
            dict1[option] = Config.get(section, option)
            if dict1[option] == -1:
                print("skip: %s" % option)
        except (NoOptionError, NoSectionError):
            print("exception on %s!" % option)
            dict1[option] = None
    return dict1


def log(message):
    print(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + ": " + message)


def get_varnames_from_filenames(files):
    if isinstance(files[0], list):
        varnames = [os.path.basename(vname[0]).split("_")[0] for vname in files]
    else:
        varnames = os.path.basename(files[0]).split("_")[0]
    return varnames


def get_input_file_path(sect_dict, infolder):
    files2list = ast.literal_eval(sect_dict["in_files"])

    # Remove infolder='opendap', this is intended as a marker
    # and not a part of the path
    infolder = re.sub("opendap", "", infolder)

    if isinstance(files2list[0], list):
        sect_dict["in_files"] = [
            [os.path.join(infolder, f) for f in fil] for fil in files2list
        ]
    else:
        sect_dict["in_files"] = [os.path.join(infolder, f) for f in files2list]
    return sect_dict


def get_time_ranges(sect_dict):
    ts_names = ["dt1", "dt2", "base_dt1", "base_dt2"]
    for dt in ts_names:
        if dt in sect_dict.keys():
            if dt == "dt1":
                dt1 = datetime.datetime.strptime(sect_dict["dt1"], "%Y-%m-%d")
                dt2 = datetime.datetime.strptime(sect_dict["dt2"], "%Y-%m-%d")
                sect_dict["time_range"] = [dt1, dt2]
                del sect_dict["dt1"]
                del sect_dict["dt2"]

            elif dt == "base_dt1":
                base_dt1 = datetime.datetime.strptime(sect_dict["base_dt1"], "%Y-%m-%d")
                base_dt2 = datetime.datetime.strptime(sect_dict["base_dt2"], "%Y-%m-%d")
                sect_dict["base_period_time_range"] = [base_dt1, base_dt2]
                del sect_dict["base_dt1"]
                del sect_dict["base_dt2"]
    return sect_dict


def get_existing_entry(dic, entry):
    if entry in dic.keys():
        return dic[entry]
    else:
        return None


def flatten_list(nested_list):
    return [item for sublist in nested_list for item in sublist]


def get_cfg(args, Config, req_section, req_option):
    if getattr(args, req_option) == "None":
        if Config.has_section(req_section):
            option = Config.get(req_section, req_option)
        else:
            option = "opendap"  # Assume opendap
    else:
        option = getattr(args, req_option)
    return option


def make_string(non_string):
    if type(non_string) is dict:
        non_string = flatten_list(list(non_string.items()))
    elif isinstance(non_string, types.FunctionType):
        non_string = non_string.__name__
    elif isinstance(non_string, tuple):
        non_string = "".join(make_string(map(list, non_string)))
    elif isinstance(non_string, datetime.datetime):
        non_string = str(non_string)
    elif isinstance(non_string, float):
        non_string = str(non_string)
    elif isinstance(non_string, int):
        non_string = str(non_string)
    elif isinstance(non_string, list):
        non_string = "".join([make_string(item) for item in non_string])
    elif isinstance(non_string, bool):
        non_string = str(non_string)
    elif non_string is None:
        non_string = str(non_string)
    else:
        pass

    if type(non_string) is not str:
        non_string = make_string(non_string)

    return non_string


def get_test_md5hash(sect_dict, *args):
    sect_dict2 = copy.deepcopy(sect_dict)
    dict_keys_str = "".join(sect_dict2.keys())
    dict_values_str = "".join([make_string(item) for item in sect_dict2.values()])

    hashid = "".join(dict_values_str) + "".join(dict_keys_str) + "".join(args)
    hashid = hashlib.md5(hashid.encode("utf-8")).hexdigest()
    return hashid


def get_output_file_path(outfolder, section, sect_dict):
    testid = get_test_md5hash(sect_dict)

    outpath = os.path.join(outfolder, "_".join([section, "__TESTID_", testid + ".nc"]))
    sect_dict["out_file"] = outpath
    return sect_dict


def try_literal_interpretation(sect_dict, key):
    if key in sect_dict.keys():
        try:
            sect_dict[key] = ast.literal_eval(sect_dict[key])
        except ValueError:
            pass
    return sect_dict


def set_none_if_noneStr(sect_dict, key):
    if key in sect_dict.keys():
        if sect_dict[key] == "None":
            sect_dict[key] = None
    return sect_dict


def get_t_range_str(t_range, name):
    if t_range is not None:
        t1_str = t_range[0].strftime("%Y%m%d")
        t2_str = t_range[1].strftime("%Y%m%d")

        return "-".join([name, t1_str, t2_str])
    else:
        return ""


def get_suffix(in_files, dt_range, bt_range):
    suffix = ""
    dt_str = get_t_range_str(dt_range, "time_range")
    bt_str = get_t_range_str(bt_range, "base_period")

    if dt_str is not None:
        suffix = suffix + dt_str

    if bt_str is not None:
        suffix = suffix + bt_str

    if isinstance(in_files[0], list):
        suffix = re.sub(
            "[0-9]{8}-[0-9]{8}.nc$",
            suffix + ".nc",
            "_".join(os.path.basename(in_files[0][0]).split("_")[1:]),
        )
    else:
        suffix = re.sub(
            "[0-9]{8}-[0-9]{8}.nc$",
            suffix + ".nc",
            "_".join(os.path.basename(in_files[0]).split("_")[1:]),
        )
    return suffix



import logging
import os
import time
from enum import Enum
from typing import Union

from icclim.icclim_exceptions import InvalidIcclimArgumentError


class Verbosity(Enum):
    LOW = ("LOW", "INFO")
    HIGH = ("HIGH", "INFO")
    SILENT = ("SILENT", "ERROR")

    def __init__(self, icc_verbosity: str, log_level: str):
        self.icc_verbosity = icc_verbosity
        self.log_level = log_level

    @staticmethod
    def lookup(query: str):
        for v in Verbosity:
            if query.upper() == v.name:
                return v
        raise InvalidIcclimArgumentError(
            f"Unrecognized log verbosity {query}. "
            f"Use one of {[v.name for v in Verbosity]}"
        )


class IcclimLogger:
    __instance = None
    verbosity: Verbosity = Verbosity.LOW
    timezone: str

    @staticmethod
    def get_instance(verbosity: Verbosity):
        if IcclimLogger.__instance is None:
            IcclimLogger(verbosity)
        return IcclimLogger.__instance

    def __init__(self, verbosity: Verbosity):
        if IcclimLogger.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            IcclimLogger.__instance = self
            self.verbosity = verbosity
            os.environ["TZ"] = "GMT"
            time.tzset()
            self.timezone = time.tzname[0]
            logging.basicConfig(
                level=verbosity.log_level, format="%(asctime)s %(message)s"
            )

    def set_verbosity(self, verbosity: Union[str, Verbosity]):
        if isinstance(verbosity, str):
            verbosity = Verbosity.lookup(verbosity)
        self.verbosity = verbosity
        logging.root.setLevel(verbosity.log_level)

    def start_message(self):
        from icclim import __version__ as icclim_version

        # flake8: noqa
        time_now = time.asctime(time.gmtime()) + " " + self.timezone
        if self.verbosity == Verbosity.SILENT:
            return
        if self.verbosity == Verbosity.LOW:
            logging.info(f"--- Icclim {icclim_version}")
            logging.info("--- BEGIN EXECUTION")
            return
        logging.info(
            "   ********************************************************************************************"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(f"   *          Icclim                {icclim_version}   *")
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            f"   *          {time_now}                                                    *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            "   *          BEGIN EXECUTION                                                                 *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            "   ********************************************************************************************"
        )

    def ending_message(self, time_cpu):
        from icclim import __version__ as icclim_version

        # flake8: noqa
        time_now = time.asctime(time.gmtime()) + " " + self.timezone
        if self.verbosity == Verbosity.SILENT:
            return
        if self.verbosity == Verbosity.LOW:
            logging.info(f"--- Icclim {icclim_version}")
            logging.info("--- CPU SECS = %-10.3f", time_cpu)
            logging.info("--- END EXECUTION")
            return
        logging.info(
            "   ********************************************************************************************"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(f"   *          Icclim                {icclim_version}   *")
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            f"   *          {time_now}                                                    *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            "   *          END EXECUTION                                                                   *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            f"   *          CP SECS = {time_cpu}                                                            *"
        )
        logging.info(
            "   *                                                                                          *"
        )
        logging.info(
            "   ********************************************************************************************"
        )

    def deprecation_warning(self, old: str, new: str):
        logging.warning(
            f"DEPRECATION_WARNING: `{old}` is deprecated. Use `{new}` instead."
        )



"""
    `icclim.models.frequency` wraps the concept of pandas frequency in order to resample
    time series.  `slice_mode` paramater of `icclim.index` is always converted to a
    `Frequency`.
"""

import datetime
from enum import Enum
from typing import Any, Callable, List, Optional, Tuple, Union

import cftime
import numpy as np
import pandas as pd
import xarray as xr
from xarray.core.dataarray import DataArray

from icclim.icclim_exceptions import InvalidIcclimArgumentError

SliceMode = Union[Any, str, List[Union[str, Tuple, int]]]


def seasons_resampler(
    month_list: List[int],
) -> Callable[[DataArray], Tuple[DataArray, DataArray]]:
    """
    Seasonal resampling method generator.
    Returns a callable of DataArray which will resample the data to
    the a season composed of the given month.
    It also attached the corresponding time_bounds.
    Parameters
    ----------
    month_list : List[int]
        List of month identified by `{1..12}`.
    Returns
    -------
    function: Callable[[DataArray], DataArray]
        function resampling the input da to the wanted season.
    """

    def resampler(da: DataArray) -> Tuple[DataArray, DataArray]:
        da_years = np.unique(da.time.dt.year)
        seasons_acc: List[DataArray] = []
        time_bounds = []
        new_time_axis = []
        start_month = month_list[0]
        end_month = month_list[-1]
        filtered_da = month_filter(da, month_list)
        # TODO, maybe raise a warning if the month_list is not made of consecutive month
        #       (case of user error)
        for year in da_years:
            if start_month > end_month:
                int_year = year - 1
            else:
                int_year = year
            first_time = filtered_da.time.values[0]
            if isinstance(first_time, cftime.datetime):
                start = cftime.datetime(
                    year, start_month, 1, calendar=first_time.calendar
                )
                end = cftime.datetime(
                    year, end_month + 1, 1, calendar=first_time.calendar
                )
            else:
                start = pd.to_datetime(f"{int_year}-{start_month}")
                end = pd.to_datetime(f"{year}-{end_month + 1}")
            end = end - datetime.timedelta(days=1)
            season = filtered_da.sel(time=slice(start, end)).sum("time")
            new_time_axis.append(start + (end - start) / 2)
            time_bounds.append([start, end])
            seasons_acc.append(season)
        seasons = xr.concat(seasons_acc, "time")
        seasons.coords["time"] = ("time", new_time_axis)
        time_bounds_da = DataArray(
            data=time_bounds,
            dims=["time", "bounds"],
            coords=[("time", seasons.time.values), ("bounds", [0, 1])],
        )
        return seasons, time_bounds_da

    return resampler


def month_filter(da: DataArray, month_list: List[int]) -> DataArray:
    return da.sel(time=da.time.dt.month.isin(month_list))


def _add_time_bounds(freq: str) -> Callable[[DataArray], Tuple[DataArray, DataArray]]:
    def add_bounds(da: DataArray) -> Tuple[DataArray, DataArray]:
        # da should already be resampled to freq
        if isinstance(da.indexes.get("time"), xr.CFTimeIndex):
            offset = xr.coding.cftime_offsets.to_offset(freq)
            start = np.array(
                [
                    cftime.datetime(
                        date.year,
                        date.month,
                        date.day,
                        date.hour,
                        date.minute,
                        date.second,
                        calendar=date.calendar,
                    )
                    for date in da.indexes.get("time")
                ]
            )
            end = start + offset
            end = end - datetime.timedelta(days=1)
        else:
            offset = pd.tseries.frequencies.to_offset(freq)
            start = pd.to_datetime(da.time.dt.floor("D"))
            end = start + offset
            end = end - pd.Timedelta(days=1)
        da["time"] = start + (end - start) / 2
        time_bounds_da = DataArray(
            data=list(zip(start, end)),
            dims=["time", "bounds"],
            coords=[("time", da.time.values), ("bounds", [0, 1])],
        )
        return da, time_bounds_da

    return add_bounds


class Frequency(Enum):
    """
    The sampling frequency of the resulting dataset.
    """

    MONTH = ("MS", ["month", "MS"], "monthly time series", _add_time_bounds("MS"))
    """ Resample to monthly values"""

    AMJJAS = (
        "MS",
        ["AMJJAS"],
        "summer half-year time series",
        seasons_resampler([*range(4, 9)]),
    )
    """ Resample to summer half-year, from April to September included."""

    ONDJFM = (
        "MS",
        ["ONDJFM"],
        "winter half-year time series",
        seasons_resampler([10, 11, 12, 1, 2, 3]),
    )
    """ Resample to winter half-year, from October to March included."""

    DJF = ("MS", ["DJF"], "winter time series", seasons_resampler([12, 1, 2]))
    """ Resample to winter season, from December to February included."""

    MAM = ("MS", ["MAM"], "spring time series", seasons_resampler([*range(3, 6)]))
    """ Resample to spring season, from March to May included."""

    JJA = ("MS", ["JJA"], "summer time series", seasons_resampler([*range(6, 9)]))
    """ Resample to summer season, from June to Agust included."""

    SON = ("MS", ["SON"], "autumn time series", seasons_resampler([*range(9, 12)]))
    """ Resample to fall season, from September to November included."""

    CUSTOM = ("MS", [], None, None)
    """ Resample to custom values. Do not use as is, use `slice_mode` with month or season
        keywords instead.
    """

    YEAR = ("YS", ["year", "YS"], "annual time series", _add_time_bounds("YS"))
    """ Resample to yearly values."""

    def __init__(
        self,
        panda_time: str,
        accepted_values: List[str],
        description: Optional[str] = None,
        post_processing: Optional[
            Callable[[DataArray], Tuple[DataArray, DataArray]]
        ] = None,
    ):
        self.panda_freq: str = panda_time
        self.accepted_values: List[str] = accepted_values
        self.description = description
        self.post_processing = post_processing

    @staticmethod
    def lookup(slice_mode: SliceMode) -> Any:
        if isinstance(slice_mode, Frequency):
            return slice_mode
        if isinstance(slice_mode, str):
            return _get_frequency_from_string(slice_mode)
        if isinstance(slice_mode, list):
            return _get_frequency_from_list(slice_mode)
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {slice_mode}."
            f"Use a Frequency from {[f for f in Frequency]}"
        )


def _get_frequency_from_string(slice_mode: str) -> Frequency:
    for freq in Frequency:
        if freq.name == slice_mode.upper() or slice_mode.upper() in map(
            str.upper, freq.accepted_values
        ):
            return freq
    raise InvalidIcclimArgumentError(f"Unknown frequency {slice_mode}.")


def _get_frequency_from_list(slice_mode_list: List) -> Frequency:
    if len(slice_mode_list) < 2:
        raise InvalidIcclimArgumentError(
            f"The given slice list {slice_mode_list}"
            f" has a length of {len(slice_mode_list)}."
            f" The maximum length here is 2."
        )
    sampling_freq = slice_mode_list[0]
    months = slice_mode_list[1]
    custom_freq = Frequency.CUSTOM
    if sampling_freq == "month":
        custom_freq.post_processing = lambda da: month_filter(da, months)
        custom_freq.description = f"monthly time series (months: {months})"
    elif sampling_freq == "season":
        if months is Tuple:
            rearranged_months = months[1] + months[0]
            custom_freq.post_processing = seasons_resampler(rearranged_months)
            custom_freq.description = (
                f"seasonal time series (season: {rearranged_months})"
            )
        else:
            custom_freq.post_processing = seasons_resampler(months)
            custom_freq.description = f"seasonal time series (season: {months})"
    else:
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {slice_mode_list}. "
            "The sampling frequency must be one of {'season', 'month'}"
        )
    return custom_freq



class IcclimError(Exception):
    """Base class for exceptions in this module."""

    pass


class InvalidIcclimArgumentError(IcclimError):
    """Exceptions raised erroneous input arguments.

    Attributes:
        msg  -- explanation of the error
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class MissingIcclimInputError(IcclimError):
    """Exceptions raised erreonous input arguments.

    Attributes:
        msg  -- explanation of the error
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class InvalidIcclimOutputError(IcclimError):
    """Exceptions raised erreonous input arguments.

    Attributes:
        msg  -- explanation of the error
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)



import pytest

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import PRECIPITATION, TEMPERATURE
from icclim.models.frequency import Frequency
from icclim.models.index_config import CfVariable
from icclim.models.user_index_config import LogicalOperation
from icclim.tests.unit_tests.test_utils import stub_pr, stub_tas, stub_user_index
from icclim.user_indices.dispatcher import CalcOperation, compute_user_index


class Test_compute:
    def test_error_bad_operation(self):
        # GIVEN
        cf_var = CfVariable(stub_tas(), stub_tas())
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = "pouet pouet"
        user_index.freq = Frequency.MONTH
        # WHEN
        with pytest.raises(InvalidIcclimArgumentError):
            compute_user_index(user_index)

    def test_simple(self):
        # GIVEN
        cf_var = CfVariable(stub_tas(), stub_tas())
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = "max"
        user_index.freq = Frequency.MONTH
        # WHEN
        result = compute_user_index(user_index)
        # THEN
        assert result.data[0] == 1

    def test_simple_percentile_pr(self):
        # GIVEN
        cf_var = CfVariable(stub_pr(5), stub_pr(5))
        cf_var.da.data[15:30] += 10
        cf_var.da.data[366 + 15 : 366 + 30] = 2  # Ignore because not in base
        cf_var.in_base_da = cf_var.da.sel(time=cf_var.da.time.dt.year == 2042)
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = CalcOperation.MIN
        user_index.thresh = "90p"
        user_index.logical_operation = LogicalOperation.GREATER_OR_EQUAL_THAN
        user_index.var_type = PRECIPITATION
        user_index.freq = Frequency.YEAR
        # WHEN
        result = compute_user_index(user_index)
        # THEN
        assert result.data[0] == 5

    def test_simple_percentile_temp(self):
        cf_var = CfVariable(stub_tas(5), stub_tas(5))
        cf_var.da.data[15:30] = 1
        cf_var.in_base_da = cf_var.da.sel(
            time=cf_var.da.time.dt.year.isin([2042, 2043])
        )
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = "min"
        user_index.thresh = "10p"
        user_index.logical_operation = LogicalOperation.LOWER_OR_EQUAL_THAN
        user_index.var_type = TEMPERATURE
        user_index.freq = Frequency.MONTH
        # WHEN
        result = compute_user_index(user_index)
        # THEN
        assert result.data[0] == 1
        assert result.data[1] == 5



from enum import Enum

from icclim.icclim_exceptions import InvalidIcclimArgumentError


class NetcdfVersion(Enum):
    NETCDF4 = "NETCDF4"
    NETCDF4_CLASSIC = "NETCDF4_CLASSIC"
    NETCDF3_CLASSIC = "NETCDF3_CLASSIC"
    NETCDF3_64BIT = "NETCDF3_64BIT"

    @staticmethod
    def lookup(s: str):
        for version in NetcdfVersion:
            if version.name.upper() == s.upper():
                return version
        raise InvalidIcclimArgumentError(f"Unknown netcdf version {s}")



from enum import Enum


class QuantileInterpolation(Enum):
    LINEAR = ("linear", 1, 1)
    MEDIAN_UNBIASED = ("hyndman_fan", 1.0 / 3, 1.0 / 3)

    def __init__(self, alias, alpha, beta):
        self.alias = alias
        self.alpha = alpha
        self.beta = beta

    @staticmethod
    def lookup(s: str):
        for interpolation in QuantileInterpolation:
            if interpolation.value.upper() == s.upper():
                return interpolation
        valid_values = list(map(lambda x: x.value, QuantileInterpolation))
        raise NotImplementedError(
            f"Interpolation must be one of the following: {valid_values}"
        )



# -*- Coding: latin-1 -*-
#  Copyright CERFACS (http://cerfacs.fr/)
#  Apache License, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0)
"""
Main module of icclim.
"""

import logging
import time
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from warnings import warn

import xarray
import xarray as xr
import xclim
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

from icclim.ecad_functions import IndexConfig
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_logger import IcclimLogger, Verbosity
from icclim.models.ecad_indices import EcadIndex
from icclim.models.frequency import Frequency, SliceMode
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation
from icclim.models.user_index_config import UserIndexConfig
from icclim.user_indices.dispatcher import compute_user_index

__all__ = ["index"]

log: IcclimLogger = IcclimLogger.get_instance(Verbosity.LOW)


def indice(*args, **kwargs):
    """
    Proxy for `index`
    To be deleted in 5.1
    """
    log.deprecation_warning(old="icclim.indice", new="icclim.index")
    return index(*args, **kwargs)


def index(
    in_files: Union[str, List[str], Dataset],
    var_name: Optional[Union[str, List[str]]] = None,
    index_name: str = None,
    slice_mode: SliceMode = Frequency.YEAR,
    time_range: List[datetime] = None,
    out_file: str = "icclim_out.nc",
    threshold: Union[float, List[float]] = None,
    transfer_limit_Mbytes: float = None,
    callback: Callable[[int], None] = lambda p: logging.info(f"Processing: {p}%"),
    callback_percentage_start_value: int = 0,
    callback_percentage_total: int = 100,
    base_period_time_range: List[datetime] = None,
    window_width: int = 5,
    only_leap_years: bool = False,
    ignore_Feb29th: bool = False,
    interpolation: Optional[
        Union[str, QuantileInterpolation]
    ] = QuantileInterpolation.MEDIAN_UNBIASED,
    out_unit: Optional[str] = None,
    netcdf_version: Union[str, NetcdfVersion] = NetcdfVersion.NETCDF4,
    # TODO do something prettier than a dict (a UserIndiceDTO or something)
    user_index: Dict[str, Any] = None,
    save_percentile: bool = False,
    logs_verbosity: Union[Verbosity, str] = Verbosity.LOW,
    # deprecated
    indice_name: str = None,
    user_indice: Dict[str, Any] = None,
) -> Dataset:
    """
    :param index_name:
        Climate index name.
    :param in_files:
        Absolute path(s) to NetCDF dataset(s) (including OPeNDAP URLs).
    :param var_name:
        Target variable name to process corresponding to ``in_files``.
    :param slice_mode:
        Type of temporal aggregation:
        {"year", "month", "DJF", "MAM", "JJA", "SON", "ONDJFM" or "AMJJAS"}.
        If ``None``, the index will be calculated as monthly values.
    :param time_range:
        Temporal range: upper and lower bounds for temporal subsetting.
        If ``None``, whole period of input files will be processed.
    :param out_file:
        Output NetCDF file name (default: "icclim_out.nc" in the current directory).
    :param threshold:
        User defined threshold for certain indices.
    :param transfer_limit_Mbytes:
        Maximum OPeNDAP/THREDDS request limit in Mbytes in case of OPeNDAP datasets.
    :param callback:
        Progress bar printing. If ``None``, progress bar will not be printed.
    :param callback_percentage_start_value:
        Initial value of percentage of the progress bar (default: 0).
    :param callback_percentage_total:
        Total persentage value (default: 100).
    :param base_period_time_range:
        Temporal range of the base period.
    :param window_width:
        Window width, must be odd (default: 5).
    :param only_leap_years:
        Option for February 29th (default: False).
    :param ignore_Feb29th:
        Ignoring or not February 29th (default: False).
    :param interpolation:
        Interpolation method to compute percentile values: "linear" or "hyndman_fan".
        default: "hyndman_fan".
    :param out_unit:
        Output unit for certain indices: "days" or "%" (default: "days").
    :param user_index:
        A dictionary with parameters for user defined index
    :param netcdf_version:
        NetCDF version to create (default: "NETCDF3_CLASSIC").
    :param save_percentile:
        True if the percentiles should be saved within the resulting netcdf file
    :rtype: path to NetCDF file

    .. warning:: If ``out_file`` already exists, icclim will overwrite it!

    """
    # make xclim input daily check a warning instead of an error
    xclim.set_options(data_validation="warn")
    # keep attributes through xarray operations
    xr.set_options(keep_attrs=True)
    log.set_verbosity(logs_verbosity)

    log.start_message()
    callback(callback_percentage_start_value)
    if indice_name is not None:
        log.deprecation_warning(old="indice_name", new="index_name")
        index_name = indice_name
    if user_indice is not None:
        log.deprecation_warning(old="user_indice", new="user_index")
        user_index = user_indice
    index: Optional[EcadIndex]
    if user_index is None:
        index = EcadIndex.lookup(index_name)
    else:
        index = None
    input_dataset, reset_coords = _read_input(in_files)
    if isinstance(var_name, str):
        var_name = [var_name]
    elif var_name is None and index is not None:
        var_name = _guess_variables(index, input_dataset)
    config = IndexConfig(
        base_period_time_range=base_period_time_range,
        ds=input_dataset,
        ignore_Feb29th=ignore_Feb29th,
        only_leap_years=only_leap_years,
        save_percentile=save_percentile,
        slice_mode=slice_mode,
        time_range=time_range,
        var_name=var_name,
        window_width=window_width,
        transfer_limit_Mbytes=transfer_limit_Mbytes,
        out_unit=out_unit,
        netcdf_version=netcdf_version,
        interpolation=interpolation,
        callback=callback,
        index=index,
    )
    if user_index is not None:
        result_ds = _compute_user_index_dataset(config, user_index)
    else:
        result_ds = _compute_ecad_index_dataset(
            config, index, threshold, input_dataset.attrs.get("history", None)
        )
    if reset_coords:
        result_ds = result_ds.rename(reset_coords)
    if not isinstance(in_files, Dataset):
        if input_dataset.time.encoding:
            if input_dataset.time.encoding.get("chunksizes"):
                del input_dataset.time.encoding["chunksizes"]
            time_encoding = input_dataset.time.encoding
        else:
            time_encoding = {"units": "days since 1850-1-1"}
        result_ds.to_netcdf(
            out_file,
            format=config.netcdf_version.value,
            encoding={"time": time_encoding},
        )
    callback(callback_percentage_total)
    log.ending_message(time.process_time())
    return result_ds


def _read_input(in_files):
    if isinstance(in_files, Dataset):
        input_dataset = in_files
    elif isinstance(in_files, list):
        input_dataset = xarray.open_mfdataset(in_files, parallel=True)
    else:
        input_dataset = xarray.open_dataset(in_files)
    input_dataset, reset_coords = _update_coords(input_dataset)
    return input_dataset, reset_coords


def _compute_ecad_index_dataset(
    config: IndexConfig,
    index: EcadIndex,
    threshold: Union[float, List[float]],
    current_history: Optional[str],
) -> Dataset:
    if isinstance(threshold, list):
        ds_list = []
        for th in threshold:
            config.threshold = th
            ds_list.append(_compute_ecad_index(index, config, current_history))
        result_ds = xarray.concat(ds_list, dim="threshold")
    else:
        config.threshold = threshold
        result_ds = _compute_ecad_index(index, config, current_history)
    return result_ds


def _compute_user_index_dataset(config: IndexConfig, user_index: dict) -> Dataset:
    logging.info("Calculating user index.")
    result_ds = Dataset()
    deprecated_name = user_index.get("indice_name", None)
    if deprecated_name is not None:
        user_index["index_name"] = deprecated_name
        del user_index["indice_name"]
        log.deprecation_warning("indice_name", "index_name")
    user_indice_config = UserIndexConfig(
        **user_index,
        freq=config.freq,
        cf_vars=config.cf_variables,
        is_percent=config.is_percent,
        save_percentile=config.save_percentile,
    )
    user_indice_da = compute_user_index(user_indice_config)
    user_indice_da.attrs["units"] = _get_unit(config.out_unit, user_indice_da)
    if config.freq.post_processing is not None:
        user_indice_da, time_bounds = config.freq.post_processing(user_indice_da)
        result_ds[user_indice_config.index_name] = user_indice_da
        result_ds["time_bounds"] = time_bounds
    else:
        result_ds[user_indice_config.index_name] = user_indice_da
    return result_ds


def _get_unit(output_unit: Optional[str], da: DataArray) -> Optional[str]:
    da_unit = da.attrs.get("units", None)
    if da_unit is None:
        if output_unit is None:
            warn(
                "No unit computed or provided for the index was found. "
                "Use out_unit parameter to add one."
            )
            return ""
        else:
            return output_unit
    else:
        return da_unit


def _compute_ecad_index(
    index: EcadIndex, config: IndexConfig, former_history: Optional[str]
) -> Dataset:
    logging.info(f"Calculating climate index: {index.index_name}")
    result_ds = Dataset()
    da, per = index.compute(config)
    da.attrs["units"] = _get_unit(config.out_unit, da)
    if config.threshold is not None:
        da.expand_dims({"threshold": config.threshold})
    if config.freq.post_processing is not None:
        resampled_da, time_bounds = config.freq.post_processing(da)
        result_ds[index.index_name] = resampled_da
        if time_bounds is not None:
            result_ds["time_bounds"] = time_bounds
            result_ds.time.attrs["bounds"] = "time_bounds"
    else:
        result_ds[index.index_name] = da
    if per is not None:
        per = per.squeeze("percentiles", drop=True).rename("percentiles")
        result_ds = xr.merge([result_ds, per])
    if former_history is None:
        former_history = da.attrs["history"]
    else:
        former_history = f"{former_history}\n{da.attrs['history']}"
    del da.attrs["history"]
    result_ds = _add_basic_indice_metadata(result_ds, config, index, former_history)
    return result_ds


def _add_basic_indice_metadata(
    result_ds: Dataset,
    config: IndexConfig,
    computed_index: EcadIndex,
    former_history: str,
) -> Dataset:
    if config.threshold is not None:
        title = f"Index {computed_index.index_name} with user defined threshold"
    else:
        title = f"ECA {computed_index.group} index {computed_index.index_name}"
    result_ds.attrs["title"] = title
    result_ds.attrs[
        "references"
    ] = "ATBD of the ECA indices calculation (https://www.ecad.eu/documents/atbd.pdf)"
    result_ds.attrs["institution"] = "Climate impact portal (https://climate4impact.eu)"
    result_ds.attrs["history"] = _get_history(
        config, former_history, computed_index, result_ds
    )
    result_ds.attrs["source"] = ""
    result_ds.attrs["Conventions"] = "CF-1.6"

    result_ds.lat.encoding["_FillValue"] = None
    result_ds.lon.encoding["_FillValue"] = None

    return result_ds


def _get_history(config, former_history, indice_computed, result_ds):
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    start_time = result_ds.time[0].dt.strftime("%m-%d-%Y").data[()]
    end_time = result_ds.time[-1].dt.strftime("%m-%d-%Y").data[()]
    return (
        f"{former_history}\n "
        f"{current_time} "
        f"Calculation of {indice_computed.index_name} "
        f"index({config.freq.description}) "
        f"from {start_time} to {end_time}."
    )


def _guess_variables(index: EcadIndex, ds: Dataset) -> List[str]:
    """
    Try to guess the variable names using the expected kind of variable for
    the index.
    """
    res = []
    index_variables = index.variables
    for indice_var in index_variables:
        for alias in indice_var:
            # check if dataset contains this alias
            if ds.get(alias, None) is not None:
                res.append(alias)
                break
    if len(res) < len(index_variables):
        variables = list(filter(lambda x: x not in ds.coords, ds.variables.keys()))
        raise InvalidIcclimArgumentError(
            f"The necessary variable(s) were not recognized in the"
            f" input file(s) to compute {index.index_name}."
            f" Use `var_name` parameter to use one the dataset non coordinate variable:"
            f" {variables}"
        )
    return res


def _update_coords(ds: Dataset) -> Tuple[Dataset, Dict]:
    revert = {}
    if ds.coords.get("latitude") is not None:
        ds = ds.rename({"latitude": "lat"})
        revert.update({"lat": "latitude"})
    if ds.coords.get("longitude") is not None:
        ds = ds.rename({"longitude": "lon"})
        revert.update({"lon": "longitude"})
    if ds.coords.get("t") is not None:
        ds = ds.rename({"t": "time"})
        revert.update({"time": "t"})
    return ds, revert



from enum import Enum
from typing import Any, List

from icclim.ecad_functions import (
    ComputeIndexFun,
    cd,
    cdd,
    cfd,
    csdi,
    csu,
    cw,
    cwd,
    dtr,
    etr,
    fd,
    gd4,
    hd17,
    id,
    prcptot,
    r10mm,
    r20mm,
    r75p,
    r75ptot,
    r95p,
    r95ptot,
    r99p,
    r99ptot,
    rr1,
    rx1day,
    rx5day,
    sd,
    sd1,
    sd5cm,
    sd50cm,
    sdii,
    su,
    tg,
    tg10p,
    tg90p,
    tn,
    tn10p,
    tn90p,
    tnn,
    tnx,
    tr,
    tx,
    tx10p,
    tx90p,
    txn,
    txx,
    vdtr,
    wd,
    wsdi,
    ww,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import (
    COLD_GROUP,
    COMPOUND_GROUP,
    DROUGHT_GROUP,
    HEAT_GROUP,
    PR,
    RAIN_GROUP,
    SNOW_GROUP,
    TAS,
    TAS_MAX,
    TAS_MIN,
    TEMPERATURE_GROUP,
)


class EcadIndex(Enum):
    """
    ECA&D indices.
        index_name: str
            The index name used in the output.
        compute: Callable
            The function to compute the index. It wraps Xclim functions.
        group: str
            The index group category. one of
            `{"temperature", "heat", "cold", "drought", "rain", "snow", "compound"}`
        variables: List[List[str]]
            The Cf variables needed to compute the index.
            The variable are individually described by a list of aliases.
    """

    # temperature
    TG = ("TG", tg, TEMPERATURE_GROUP, [TAS])
    TN = ("TN", tn, TEMPERATURE_GROUP, [TAS_MIN])
    TX = ("TX", tx, TEMPERATURE_GROUP, [TAS_MAX])
    DTR = ("DTR", dtr, TEMPERATURE_GROUP, [TAS_MAX, TAS_MIN])
    ETR = ("ETR", etr, TEMPERATURE_GROUP, [TAS_MAX, TAS_MIN])
    VDTR = ("vDTR", vdtr, TEMPERATURE_GROUP, [TAS_MAX, TAS_MIN])
    # heat
    SU = ("SU", su, HEAT_GROUP, [TAS_MAX])
    TR = ("TR", tr, HEAT_GROUP, [TAS_MIN])
    WSDI = ("WSDI", wsdi, HEAT_GROUP, [TAS_MAX])
    TG90P = ("TG90p", tg90p, HEAT_GROUP, [TAS])
    TN90P = ("TN90p", tn90p, HEAT_GROUP, [TAS_MIN])
    TX90P = ("TX90p", tx90p, HEAT_GROUP, [TAS_MAX])
    TXX = ("TXx", txx, HEAT_GROUP, [TAS_MAX])
    TNX = ("TNx", tnx, HEAT_GROUP, [TAS_MIN])
    CSU = ("CSU", csu, HEAT_GROUP, [TAS_MAX])
    # cold
    GD4 = ("GD4", gd4, COLD_GROUP, [TAS])
    FD = ("FD", fd, COLD_GROUP, [TAS_MIN])
    CFD = ("CFD", cfd, COLD_GROUP, [TAS_MIN])
    HD17 = ("HD17", hd17, COLD_GROUP, [TAS])
    ID = ("ID", id, COLD_GROUP, [TAS_MAX])
    TG10P = ("TG10p", tg10p, COLD_GROUP, [TAS])
    TN10P = ("TN10p", tn10p, COLD_GROUP, [TAS_MIN])
    TX10P = ("TX10p", tx10p, COLD_GROUP, [TAS_MAX])
    TXN = ("TXn", txn, COLD_GROUP, [TAS_MAX])
    TNN = ("TNn", tnn, COLD_GROUP, [TAS_MIN])
    CSDI = ("CSDI", csdi, COLD_GROUP, [TAS_MIN])
    # drought
    CDD = ("CDD", cdd, DROUGHT_GROUP, [PR])
    # rain
    PRCPTOT = ("PRCPTOT", prcptot, RAIN_GROUP, [PR])
    RR1 = ("RR1", rr1, RAIN_GROUP, [PR])
    SDII = ("SDII", sdii, RAIN_GROUP, [PR])
    CWD = ("CWD", cwd, RAIN_GROUP, [PR])
    R10MM = ("R10mm", r10mm, RAIN_GROUP, [PR])
    R20MM = ("R20mm", r20mm, RAIN_GROUP, [PR])
    RX1DAY = ("RX1day", rx1day, RAIN_GROUP, [PR])
    RX5DAY = ("RX5day", rx5day, RAIN_GROUP, [PR])
    R75P = ("R75p", r75p, RAIN_GROUP, [PR])
    R75PTOT = ("R75pTOT", r75ptot, RAIN_GROUP, [PR])
    R95P = ("R95p", r95p, RAIN_GROUP, [PR])
    R95PTOT = ("R95pTOT", r95ptot, RAIN_GROUP, [PR])
    R99P = ("R99p", r99p, RAIN_GROUP, [PR])
    R99PTOT = ("R99pTOT", r99ptot, RAIN_GROUP, [PR])
    # snow
    SD = ("SD", sd, SNOW_GROUP, [PR])
    SD1 = ("SD1", sd1, SNOW_GROUP, [PR])
    SD5CM = ("SD5cm", sd5cm, SNOW_GROUP, [PR])
    SD50CM = ("SD50cm", sd50cm, SNOW_GROUP, [PR])
    # compound
    CD = ("CD", cd, COMPOUND_GROUP, [TAS, PR])
    CW = ("CW", cw, COMPOUND_GROUP, [TAS, PR])
    WD = ("WD", wd, COMPOUND_GROUP, [TAS, PR])
    WW = ("WW", ww, COMPOUND_GROUP, [TAS, PR])

    def __init__(
        self,
        index_name: str,
        compute: ComputeIndexFun,
        group: str,
        variables: List[List[str]],
    ):
        self.index_name = index_name
        self.compute = compute
        self.group = group
        self.variables = variables

    @staticmethod
    def lookup(query: str) -> Any:
        for e in EcadIndex:
            if e.index_name.upper() == query.upper():
                return e
        raise InvalidIcclimArgumentError(f"Unknown index {query}.")



import numpy as np
import pytest

from icclim.models.user_index_config import (
    ExtremeMode,
    LinkLogicalOperation,
    LogicalOperation,
)
from icclim.tests.unit_tests.test_utils import stub_tas
from icclim.user_indices.operators import (
    _apply_coef,
    anomaly,
    count_events,
    max,
    max_consecutive_event_count,
    mean,
    min,
    run_mean,
    run_sum,
    sum,
)


class Test_apply_coef:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        # GIVEN
        da = stub_tas(use_dask=use_dask)
        # WHEN
        result = _apply_coef(4.0, da)
        # THEN
        assert np.testing.assert_equal(result.data, 4.0) is None


class Test_max:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        da = stub_tas(use_dask=use_dask)
        da.data[1] = 20
        # WHEN
        result = max(
            da=da,
            coef=1,
            logical_operation=None,
            threshold=None,
            freq="YS",
            date_event=True,
        )
        # THEN
        assert result.data[0] == 20


class Test_min:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        da = stub_tas(use_dask=use_dask)
        da.data[1] = -20
        # WHEN
        result = min(
            da=da,
            freq="YS",
        )
        # THEN
        assert result.data[0] == -20


class Test_mean:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        da = stub_tas(use_dask=use_dask)
        da[2] = 366
        # WHEN
        result = mean(
            da=da,
            freq="YS",
        )
        # THEN
        assert result.data[0] == 2


class Test_sum:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        da = stub_tas(use_dask=use_dask)
        # WHEN
        result = sum(
            da=da,
            freq="YS",
        )
        # THEN
        assert result.data[0] == 365


class Test_count_events:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        # GIVEN
        da = stub_tas(10, use_dask)
        da[1] = 15
        da[2] = 16
        # WHEN
        result = count_events(
            das=[da],
            in_base_das=[None],
            logical_operation=[LogicalOperation.GREATER_THAN],
            thresholds=[15],
            freq="MS",
        )
        # THEN
        assert result[0] == 1

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_percentile(self, use_dask):
        # GIVEN
        da = stub_tas(10, use_dask)
        da[1] = 15
        da[2] = 16
        # WHEN
        result = count_events(
            das=[da],
            in_base_das=[da],
            logical_operation=[LogicalOperation.GREATER_THAN],
            thresholds=["80p"],
            freq="MS",
        )
        # THEN
        assert result[0] == 2

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_multi_threshold_or(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[1] = 15
        tmin = stub_tas(-10, use_dask)
        # WHEN
        result = count_events(
            das=[tmax, tmin],
            in_base_das=[None],
            logical_operation=[LogicalOperation.GREATER_THAN, LogicalOperation.EQUAL],
            thresholds=[12, -20],
            link_logical_operations=LinkLogicalOperation.OR_STAMP,
            freq="MS",
        )
        # THEN
        assert result[0] == 1

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_multi_threshold_and(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[1] = 15
        tmin = stub_tas(-10, use_dask)
        tmin[1] = -20
        # WHEN
        result = count_events(
            das=[tmax, tmin],
            in_base_das=[None],
            logical_operation=[LogicalOperation.GREATER_THAN, LogicalOperation.EQUAL],
            thresholds=[12, -20],
            link_logical_operations=LinkLogicalOperation.AND_STAMP,
            freq="MS",
        )
        # THEN
        assert result[0] == 1


class Test_run_mean:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_min(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[30] = 0
        tmax[29] = 0
        tmax[28] = 0
        tmax[27] = 0
        tmax[26] = 0
        # WHEN
        result = run_mean(
            da=tmax,
            extreme_mode=ExtremeMode.MIN,
            window_width=5,
            freq="MS",
        )
        # THEN
        assert result[0] == 0
        assert result[1] == 2
        assert result[2] == 10

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_max(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[30] = 20
        # WHEN
        result = run_mean(
            da=tmax,
            extreme_mode=ExtremeMode.MAX,
            window_width=2,
            freq="MS",
        )
        # THEN
        assert result[0] == 15
        assert result[1] == 15
        assert result[2] == 10


class Test_run_sum:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_min(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[30] = 0
        tmax[29] = 0
        tmax[28] = 0
        tmax[27] = 0
        tmax[26] = 0
        # WHEN
        result = run_sum(
            da=tmax,
            extreme_mode=ExtremeMode.MIN,
            window_width=5,
            freq="MS",
        )
        # THEN
        assert result[0] == 0
        assert result[1] == 10
        assert result[2] == 50

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_max(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[30] = 20
        # WHEN
        result = run_sum(
            da=tmax,
            extreme_mode=ExtremeMode.MAX,
            window_width=2,
            freq="MS",
        )
        # THEN
        assert result[0] == 30
        assert result[1] == 30
        assert result[2] == 20


class Test_max_consecutive_event_count:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax[30] = 15  # On 31th january
        # WHEN
        result = max_consecutive_event_count(
            da=tmax,
            logical_operation=LogicalOperation.EQUAL,
            threshold=10.0,
            freq="YS",
        )
        # THEN
        assert result[0] == 334
        assert result[1] == 365


class Test_anomaly:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax2 = stub_tas(11, use_dask)
        # WHEN
        result = anomaly(da_ref=tmax, da=tmax2, percent=False)
        # THEN
        assert result == 1

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_percent(self, use_dask):
        # GIVEN
        tmax = stub_tas(10, use_dask)
        tmax2 = stub_tas(11, use_dask)
        # WHEN
        result = anomaly(da_ref=tmax, da=tmax2, percent=True)
        # THEN
        assert result == 10
        assert result.attrs["units"] == "%"



from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, List, Optional, Union

from xarray.core.dataarray import DataArray

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.frequency import Frequency
from icclim.models.index_config import CfVariable


class LinkLogicalOperation(Enum):
    OR_STAMP = "or"
    AND_STAMP = "and"

    @staticmethod
    def lookup(s: str) -> Any:
        for mode in LinkLogicalOperation:
            if s.upper == mode.value.upper():
                return mode
        raise InvalidIcclimArgumentError(
            f"Unknown link_logical_operation mode {s}."
            f"Use one of {[linkOp.value for linkOp in LinkLogicalOperation]}."
        )


class ExtremeMode(Enum):
    MIN = "min"
    MAX = "max"

    @staticmethod
    def lookup(s: str) -> Any:
        for mode in ExtremeMode:
            if s.upper == mode.value.upper():
                return mode
        raise InvalidIcclimArgumentError(
            f"Unknown extreme mode {s}."
            f"Use one of {[mode.value for mode in ExtremeMode]}."
        )


class LogicalOperation(Enum):
    GREATER_THAN = (["gt", ">"], ">", lambda da, th: da > th)
    LOWER_THAN = (["lt", "<"], "<", lambda da, th: da < th)
    GREATER_OR_EQUAL_THAN = (["get", "ge", ">=", "=>"], ">=", lambda da, th: da >= th)
    LOWER_OR_EQUAL_THAN = (["let", "le", "<=", "=<"], "<=", lambda da, th: da <= th)
    EQUAL = (["e", "equal", "eq", "=", "=="], "==", lambda da, th: da == th)

    def __init__(
        self,
        aliases: str,
        operator: str,
        compute: Callable[[DataArray, Union[DataArray, float, int]], DataArray],
    ) -> None:
        super().__init__()
        self.aliases = aliases
        self.operator = operator
        self.compute = compute

    @staticmethod
    def lookup(s: str) -> Any:
        for op in LogicalOperation:
            if s.upper() in map(str.upper, op.aliases):
                return op
        raise InvalidIcclimArgumentError(
            f"Unknown logical operator {s}."
            f"Use one of {[op.aliases for op in LogicalOperation]}."
        )


@dataclass
class NbEventConfig:
    logical_operation: List[LogicalOperation]
    thresholds: List[Union[float, str]]
    link_logical_operations: Optional[LinkLogicalOperation] = None
    data_arrays: Optional[List[CfVariable]] = None


@dataclass
class UserIndexConfig:
    index_name: str
    calc_operation: str
    cf_vars: List[CfVariable]
    freq: Frequency
    date_event: bool
    is_percent: bool
    logical_operation: Optional[LogicalOperation] = None
    thresh: Optional[Union[float, int, str, List[Union[float, int, str]]]] = None
    link_logical_operations: Optional[LinkLogicalOperation] = None
    extreme_mode: Optional[ExtremeMode] = None
    window_width: Optional[int] = None
    coef: Optional[float] = None
    var_type: Optional[str] = None
    da_ref: Optional[DataArray] = None
    nb_event_config: Optional[NbEventConfig] = None
    save_percentile: bool = False

    def __init__(
        self,
        index_name,
        # Any should be CalcOperation but it causes circular import
        calc_operation: Union[str, Any],
        freq: Frequency,
        cf_vars: List[CfVariable],
        logical_operation: str = None,
        thresh=None,
        link_logical_operations: str = None,
        extreme_mode: str = None,
        window_width=None,
        coef=None,
        date_event=None,
        var_type=None,
        is_percent=False,
        save_percentile=False,
    ) -> None:
        self.index_name = index_name
        self.calc_operation = calc_operation
        self.freq = freq
        if logical_operation is not None:
            self.logical_operation = LogicalOperation.lookup(logical_operation)
        self.thresh = thresh
        if extreme_mode is not None:
            self.extreme_mode = ExtremeMode.lookup(extreme_mode)
        self.window_width = window_width
        self.coef = coef
        self.date_event = date_event
        self.var_type = var_type
        self.is_percent = is_percent
        self.da_ref = cf_vars[0].in_base_da
        self.cf_vars = cf_vars
        if thresh is not None and logical_operation is not None:
            self.nb_event_config = get_nb_event_conf(
                logical_operation, link_logical_operations, thresh, cf_vars
            )
        self.save_percentile = save_percentile


def get_nb_event_conf(
    logical_operation: Union[List[str], str],
    link_logical_operations: Optional[str],
    thresholds: Union[List[Union[str, float]], float, str],
    cfvars: List[CfVariable],
) -> NbEventConfig:
    if not isinstance(thresholds, list):
        threshold_list = [thresholds]
    else:
        threshold_list = thresholds
    if isinstance(logical_operation, list):
        logical_operations = list(map(LogicalOperation.lookup, logical_operation))
    else:
        logical_operations = [LogicalOperation.lookup(logical_operation)]
    if link_logical_operations is not None:
        link_logical_operation_list = LinkLogicalOperation.lookup(
            link_logical_operations
        )
    else:
        link_logical_operation_list = None
    return NbEventConfig(
        logical_operation=logical_operations,
        link_logical_operations=link_logical_operation_list,
        thresholds=threshold_list,
        data_arrays=cfvars,
    )



from enum import Enum
from typing import Any, Callable, List

import numpy as np
import xarray as xr
from xarray import DataArray


def _proleptic_gregorian_leap(years: DataArray) -> DataArray:
    return np.logical_or(
        years % 400 == 0, np.logical_and(years % 100 != 0, years % 4 == 0)
    )


def _julian_leap(years: DataArray) -> DataArray:
    return years % 4


def _standard_leap(years: DataArray) -> DataArray:
    res = xr.full_like(years, 0)
    res[years < 1582] = _julian_leap(years)
    res[years >= 1582] = _proleptic_gregorian_leap(years)
    return res


class CfCalendar(Enum):
    """
    CF defined calendars with some additional aliases names for convenience.
    The first value of the aliases is the calendar "main" name.

    aliases: List[str]
        All the possible aliases or poorly typed calendar names targeting the same
        calendar.
    is_leap: Callable
        It expects a DataArray argument of years such as `da.time.dt.year`.
        Returns a mask of the input telling if the value is part of a leap year or not.

    """

    NO_LEAP = (
        ["noleap", "no_leap", "days_365", "days365", "365_day", "365day"],
        lambda da: xr.full_like(da, False).values,
    )
    DAYS_360 = (
        ["360_day", "days_360", "360day", "days360"],
        lambda da: xr.full_like(da, False).values,
    )
    ALL_LEAP = (
        ["all_leap", "allleap", "days_366", "days366", "366_day", "366day"],
        lambda da: xr.full_like(da, True).values,
    )
    PROLEPTIC_GREGORIAN = (
        ["proleptic_gregorian", "prolepticgregorian"],
        lambda da: _proleptic_gregorian_leap(da).values,
    )
    JULIAN = (["julian"], lambda da: _julian_leap(da).values)
    STANDARD = (["standard", "gregorian"], lambda da: _standard_leap(da).values)
    # Not sure what to do with none calendar
    NONE = (["none"], lambda da: _standard_leap(da).values)

    def __init__(self, aliases: List[str], is_leap: Callable[[DataArray], np.ndarray]):
        self.aliases = aliases
        self.is_leap = is_leap

    def get_name(self) -> str:
        return self.aliases[0]

    @staticmethod
    def lookup(query: str) -> Any:
        res = list(filter(lambda x: query.lower() in x.aliases, CfCalendar))
        if len(res) == 0:
            raise TypeError(f"No calendars found for query '{query}'")
        return res[0]



import datetime

import numpy as np
import pytest
from xarray import Dataset

from icclim.ecad_functions import cfd, csu, fd, gd4, hd17, prcptot, su, tn10p, tr, tx90p
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.ecad_indices import EcadIndex
from icclim.models.frequency import Frequency
from icclim.models.index_config import IndexConfig
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation
from icclim.tests.unit_tests.test_utils import K2C, stub_pr, stub_tas


class Test_index_from_string:
    def test_simple(self):
        res = EcadIndex.lookup("SU")
        assert res == EcadIndex.SU

    def test_lowercase(self):
        res = EcadIndex.lookup("tx90p")
        assert res == EcadIndex.TX90P

    def test_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            EcadIndex.lookup("cacahuête")


@pytest.mark.parametrize("use_dask", [True, False])
def test_tn10p_interpolation_error(use_dask):
    ds = Dataset()
    ds["tas"] = stub_tas(use_dask=use_dask)
    conf = IndexConfig(
        ds=ds,
        slice_mode=Frequency.MONTH,
        var_name=["tas"],
        index=EcadIndex.TN10P,
        netcdf_version=NetcdfVersion.NETCDF4,
        base_period_time_range=[
            ds.time.values[0].astype("M8[D]").astype("O"),
            ds.time.values[-1].astype("M8[D]").astype("O"),
        ],
        window_width=2,
        interpolation=QuantileInterpolation.LINEAR,
    )
    with pytest.raises(InvalidIcclimArgumentError):
        tn10p(conf)


@pytest.mark.parametrize("use_dask", [True, False])
def test_tn10p(use_dask):
    ds = Dataset()
    ds["tas"] = stub_tas(use_dask=use_dask)
    conf = IndexConfig(
        ds=ds,
        slice_mode=Frequency.MONTH,
        var_name=["tas"],
        netcdf_version=NetcdfVersion.NETCDF4,
        base_period_time_range=[
            ds.time.values[0].astype("M8[D]").astype("O"),
            ds.time.values[-1].astype("M8[D]").astype("O"),
        ],
        window_width=2,
        interpolation=QuantileInterpolation.MEDIAN_UNBIASED,
        save_percentile=True,
        index=EcadIndex.TN10P,
    )
    res = tn10p(conf)
    assert res is not None


class Test_SU:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_su_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[:5] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.SU,
        )
        res = su(conf)
        assert res is not None
        assert res[0][0] == 26  # January

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_su_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(use_dask=use_dask)
        ds.tas[:5] = 50 + K2C
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=40,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.SU,
        )
        res = su(conf)
        assert res is not None
        assert res[0][0] == 5  # January


class Test_TR:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[:5] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.TR,
        )
        res = tr(conf)
        assert res is not None
        assert res[0][0] == 26  # January

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(use_dask=use_dask)
        ds.tas[:5] = 50 + K2C
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=40,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.TR,
        )
        res = tr(conf)
        assert res is not None
        assert res[0][0] == 5  # January


class Test_prcptot:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["pr"] = stub_pr(value=2, use_dask=use_dask)
        ds.pr[:10] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["pr"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.PRCPTOT,
        )
        res = prcptot(conf)
        assert res is not None
        np.testing.assert_almost_equal(res[0][0], 42.0, 14)


class Test_csu:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[10:15] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.CSU,
        )
        res = csu(conf)
        assert res is not None
        assert res[0][0] == 16  # January

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(use_dask=use_dask)
        ds.tas[:5] = 50 + K2C
        ds.tas[10:20] = 50 + K2C
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=40,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.CSU,
        )
        res = csu(conf)
        assert res is not None
        assert res[0][0] == 10  # January


class Test_gd4:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[5:15] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.GD4,
        )
        res = gd4(conf)
        assert res is not None
        expected = (26 - 4) * 21
        assert res[0][0] == expected  # 21 days in January above 4 degC (at 26degC)

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[5:15] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=5,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.GD4,
        )
        res = gd4(conf)
        assert res is not None
        expected = (26 - 5) * 21
        assert res[0][0] == expected  # 21 days in January above 4 degC (at 26degC)


class Test_cfd:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[5:15] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.CFD,
        )
        res = cfd(conf)
        assert res is not None
        assert res[0][0] == 10

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[5:10] = 0
        ds.tas[10:15] = 4
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=5,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.CFD,
        )
        res = cfd(conf)
        assert res is not None
        assert res[0][0] == 10


class Test_fd:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[5:15] = 0
        ds.tas[20:25] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.FD,
        )
        res = fd(conf)
        assert res is not None
        assert res[0][0] == 15

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=26 + K2C, use_dask=use_dask)
        ds.tas[5:10] = 0
        ds.tas[10:15] = 4
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=5,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.FD,
        )
        res = fd(conf)
        assert res is not None
        assert res[0][0] == 10


class Test_hd17:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_default_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=27 + K2C, use_dask=use_dask)
        ds.tas[5:10] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.HD17,
        )
        res = hd17(conf)
        assert res is not None
        assert res[0][0] == 5 * (17 + K2C)

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_custom_threshold(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=27 + K2C, use_dask=use_dask)
        ds.tas[5:10] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            threshold=5,
            netcdf_version=NetcdfVersion.NETCDF4,
            index=EcadIndex.HD17,
        )
        res = hd17(conf)
        assert res is not None
        assert res[0][0] == 5 * (5 + K2C)


class TestTx90p:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_no_bootstrap_no_overlap(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=27 + K2C, use_dask=use_dask)
        ds.tas[5:10] = 0
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            base_period_time_range=[
                datetime.datetime(2042, 1, 1),
                datetime.datetime(2042, 12, 31),
            ],
            time_range=[datetime.datetime(2043, 1, 1), datetime.datetime(2045, 12, 31)],
            index=EcadIndex.TX90P,
        )
        res, _ = tx90p(conf)
        assert "reference_epoch" not in res.attrs.keys()

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_no_bootstrap_1_year_base(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=27 + K2C, use_dask=use_dask)
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            base_period_time_range=[
                datetime.datetime(2042, 1, 1),
                datetime.datetime(2042, 12, 31),
            ],
            time_range=[datetime.datetime(2042, 1, 1), datetime.datetime(2045, 12, 31)],
            index=EcadIndex.TX90P,
        )
        res, _ = tx90p(conf)
        assert "reference_epoch" not in res.attrs.keys()

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_bootstrap_2_years(self, use_dask):
        ds = Dataset()
        ds["tas"] = stub_tas(value=27 + K2C, use_dask=use_dask)
        conf = IndexConfig(
            ds=ds,
            slice_mode=Frequency.MONTH,
            var_name=["tas"],
            netcdf_version=NetcdfVersion.NETCDF4,
            base_period_time_range=[
                datetime.datetime(2042, 1, 1),
                datetime.datetime(2043, 12, 31),
            ],
            time_range=[datetime.datetime(2042, 1, 1), datetime.datetime(2045, 12, 31)],
            index=EcadIndex.TX90P,
        )
        res, _ = tx90p(conf)
        assert res.attrs["reference_epoch"] == ["2042-01-01", "2043-12-31"]



import numpy as np
import pandas as pd
import pytest

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.frequency import Frequency, month_filter, seasons_resampler
from icclim.tests.unit_tests.test_utils import stub_tas


class Test_build_frequency_over_frequency:
    def test_simple(self):
        freq = Frequency.lookup(Frequency.YEAR)
        assert freq == Frequency.YEAR


class Test_build_frequency_over_string:
    def test_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            Frequency.lookup("yolo")

    def test_simple(self):
        freq = Frequency.lookup("year")
        assert freq == Frequency.YEAR


class Test_build_frequency_over_list:
    def test_error(self):
        with pytest.raises(InvalidIcclimArgumentError):
            Frequency.lookup(["cacahuêtes"])

    def test_month(self):
        freq = Frequency.lookup(["month", [1, 4, 3]])
        assert freq == Frequency.CUSTOM
        assert freq.panda_freq == "MS"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_season(self):
        freq = Frequency.lookup(["season", [1, 2, 3, 4]])
        assert freq == Frequency.CUSTOM
        assert freq.panda_freq == "MS"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_winter_deprecated(self):
        # deprecated way
        freq = Frequency.lookup(["season", ([11, 12], [3, 4])])
        assert freq == Frequency.CUSTOM
        assert freq.panda_freq == "MS"
        assert freq.accepted_values == []
        assert freq.post_processing is not None

    def test_winter(self):
        freq = Frequency.lookup(["season", [11, 12, 1, 2]])
        assert freq == Frequency.CUSTOM
        assert freq.panda_freq == "MS"
        assert freq.accepted_values == []
        assert freq.post_processing is not None


class Test_filter_months:
    def test_simple(self):
        # WHEN
        da = month_filter(stub_tas(), [1, 2, 7])
        # THEN
        months = np.unique(da.time.dt.month)
        assert len(months) == 3
        assert months[0] == 1
        assert months[1] == 2
        assert months[2] == 7


class Test_seasons_resampler:
    def test_simple(self):
        # WHEN
        da_res, time_bds_res = seasons_resampler([4, 5, 6])(stub_tas())
        # THEN
        assert da_res[0] == 91
        assert time_bds_res[0].data[0] == pd.to_datetime("2042-04")
        assert (
            time_bds_res[0].data[1]
            == pd.to_datetime("2042-07") - pd.tseries.offsets.Day()
        )

    def test_winter(self):
        # WHEN
        da_res, time_bds_res = seasons_resampler([11, 12, 1])(stub_tas())
        # THEN
        assert da_res[0] == 31
        assert time_bds_res[0].data[0] == pd.to_datetime("2041-11")
        assert (
            time_bds_res[0].data[1]
            == pd.to_datetime("2042-02") - pd.tseries.offsets.Day()
        )
        assert da_res[1] == 92
        assert time_bds_res[1].data[0] == pd.to_datetime("2042-11")
        assert (
            time_bds_res[1].data[1]
            == pd.to_datetime("2043-02") - pd.tseries.offsets.Day()
        )

    def test_season_with_holes(self):
        # WHEN
        da_res, time_bds_res = seasons_resampler([1, 3, 4])(stub_tas())
        # THEN
        assert da_res[0] == 92
        assert time_bds_res[0].data[0] == pd.to_datetime("2042-01")
        assert (
            time_bds_res[0].data[1]
            == pd.to_datetime("2042-05") - pd.tseries.offsets.Day()
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

from icclim import __version__

MINIMAL_REQUIREMENTS = [
    "numpy>=1.16",
    "xarray>=0.17",
    "xclim>=0.31",
    "scipy>=1.2",
    "cftime>=1.4.1",
    "dask[array]>=2.6",
    "netCDF4>=1.5.7",
]

setup(
    name="icclim",
    version=__version__,
    packages=find_packages(),
    author="Christian P.",
    author_email="christian.page@cerfacs.fr",
    description="Python library for climate indices calculation",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    include_package_data=True,
    url="https://github.com/cerfacs-globc/icclim",
    install_requires=MINIMAL_REQUIREMENTS,
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: French",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
)



from typing import List

import numpy as np
import pandas as pd
import xarray

from icclim.models.frequency import Frequency
from icclim.models.index_config import CfVariable
from icclim.models.user_index_config import UserIndexConfig

VALUE_COUNT = 365 * 5 + 1
COORDS = dict(
    lat=[42],
    lon=[42],
    time=pd.date_range("2042-01-01", periods=VALUE_COUNT, freq=pd.DateOffset(days=1)),
)
K2C = 273.15


def stub_user_index(cf_vars: List[CfVariable]):
    return UserIndexConfig(
        index_name="Yolo", calc_operation="noop", freq=Frequency.MONTH, cf_vars=cf_vars
    )


def stub_tas(value: float = 1, use_dask=False):
    da = xarray.DataArray(
        data=(np.full(VALUE_COUNT, value).reshape((VALUE_COUNT, 1, 1))),
        dims=["time", "lat", "lon"],
        coords=COORDS,
        attrs={"units": "K"},
    )
    if use_dask:
        da.chunk()
    return da


def stub_pr(value: float, use_dask=False):
    da = xarray.DataArray(
        data=np.full(VALUE_COUNT, value).reshape((VALUE_COUNT, 1, 1)),
        coords=COORDS,
        dims=["time", "lat", "lon"],
        name="pr",
        attrs={"units": "kg m-2 d-1"},
    )
    if use_dask:
        da.chunk()
    return da



import pytest

from icclim.models.frequency import Frequency
from icclim.models.index_config import CfVariable
from icclim.models.user_index_config import LogicalOperation, UserIndexConfig
from icclim.tests.unit_tests.test_utils import stub_tas


class Test_UserindexConfig:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple_from_dict(self, use_dask):
        dico = {
            "index_name": "my_index",
            "calc_operation": "min",
            "logical_operation": "gt",
            "thresh": 0 + 273.15,
            "date_event": True,
        }
        tas = stub_tas(use_dask=use_dask)
        config = UserIndexConfig(
            **dico, freq=Frequency.MONTH, cf_vars=[CfVariable(tas, tas)]
        )
        assert config.index_name == "my_index"
        assert config.calc_operation == "min"
        assert config.logical_operation == LogicalOperation.GREATER_THAN
        assert config.thresh == 273.15
        assert config.date_event
        assert config.freq == Frequency.MONTH
        assert config.cf_vars[0].da is tas



"""
All ECA&D functions. Each function wraps its Xclim equivalent functions adding icclim
metadata to it.
"""
from typing import Callable, Optional, Tuple, Union
from warnings import warn

import numpy as np
import xarray as xr
import xclim.core.utils
from xarray import DataArray
from xclim import atmos, land
from xclim.core.calendar import percentile_doy, resample_doy
from xclim.core.units import convert_units_to

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.cf_calendar import CfCalendar
from icclim.models.constants import IN_BASE_IDENTIFIER, PERCENTILES_COORD
from icclim.models.frequency import Frequency
from icclim.models.index_config import IndexConfig
from icclim.models.quantile_interpolation import QuantileInterpolation

ComputeIndexFun = Callable[[IndexConfig], Tuple[DataArray, Optional[DataArray]]]


def gd4(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "4.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.growing_degree_days(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def cfd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "0.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.consecutive_frost_days(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def fd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "0.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.frost_days(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def hd17(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "17.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.heating_degree_days(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def id(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "0.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.ice_days(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def csdi(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 10
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.cold_spell_duration_index(
        config.cf_variables[0].da,
        per,
        window=6,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.save_percentile:
        return result, per
    return result, None


def tg10p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 10
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.tg10p(
        config.cf_variables[0].da,
        per,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def tn10p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 10
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.tn10p(
        config.cf_variables[0].da,
        per,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def tx10p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 10
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.tx10p(
        config.cf_variables[0].da,
        per,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def txn(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tx_min(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def tnn(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tn_min(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def cdd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.maximum_consecutive_dry_days(
        config.cf_variables[0].da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result, None


def su(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "25.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.tx_days_above(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def tr(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "20.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.tropical_nights(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def wsdi(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 90
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.warm_spell_duration_index(
        config.cf_variables[0].da,
        per,
        window=6,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.save_percentile:
        return result, per
    return result, None


def tg90p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 90
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.tg90p(
        config.cf_variables[0].da,
        per,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def tn90p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 90
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.tn90p(
        config.cf_variables[0].da,
        per,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def tx90p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    per_value = 90
    per = _compute_percentile_doy(config.cf_variables[0].in_base_da, config, per_value)
    run_bootstrap = _can_run_bootstrap(config, slice(*per.climatology_bounds))
    result = atmos.tx90p(
        config.cf_variables[0].da,
        per,
        freq=config.freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def txx(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tx_max(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def tnx(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tn_max(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def csu(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    if config.threshold is None:
        threshold = "25.0 degC"
    else:
        threshold = _add_celsius_suffix(config.threshold)
    result = atmos.maximum_consecutive_warm_days(
        config.cf_variables[0].da, thresh=threshold, freq=config.freq.panda_freq
    )
    return result, None


def prcptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.precip_accumulation(
        _filter_in_wet_days(config.cf_variables[0].da, dry_day_value=0),
        freq=config.freq.panda_freq,
    )
    return result, None


def rr1(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.wetdays(
        config.cf_variables[0].da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result, None


def sdii(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.daily_pr_intensity(
        config.cf_variables[0].da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result, None


def cwd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.maximum_consecutive_wet_days(
        config.cf_variables[0].da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result, None


def r10mm(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.wetdays(
        config.cf_variables[0].da, thresh="10 mm/day", freq=config.freq.panda_freq
    )
    return result, None


def r20mm(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.wetdays(
        config.cf_variables[0].da, thresh="20 mm/day", freq=config.freq.panda_freq
    )
    return result, None


def rx1day(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.max_1day_precipitation_amount(
        config.cf_variables[0].da, freq=config.freq.panda_freq
    )
    return result, None


def rx5day(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.max_n_day_precipitation_amount(
        config.cf_variables[0].da, window=5, freq=config.freq.panda_freq
    )
    return result, None


def r75p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(
        config.cf_variables[0].in_base_da, dry_day_value=np.nan
    )
    per = _compute_percentile(base_wet_days, config, 75.0)
    result = atmos.days_over_precip_thresh(
        config.cf_variables[0].da,
        per,
        thresh="1 mm/day",
        freq=config.freq.panda_freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def r75ptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(
        config.cf_variables[0].in_base_da, dry_day_value=np.nan
    )
    per = _compute_percentile(base_wet_days, config, 75.0)
    result = atmos.fraction_over_precip_thresh(
        config.cf_variables[0].da,
        per,
        thresh="1 mm/day",
        freq=config.freq.panda_freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    result = result * 100
    result.attrs["units"] = "%"
    if config.save_percentile:
        return result, per
    return result, None


def r95p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(
        config.cf_variables[0].in_base_da, dry_day_value=np.nan
    )
    per = _compute_percentile(base_wet_days, config, 95.0)
    result = atmos.days_over_precip_thresh(
        config.cf_variables[0].da,
        per,
        thresh="1 mm/day",
        freq=config.freq.panda_freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if config.is_percent:
        result = _to_percent(result, config.freq)
    if config.save_percentile:
        return result, per
    return result, None


def r95ptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(
        config.cf_variables[0].in_base_da, dry_day_value=np.nan
    )
    per = _compute_percentile(base_wet_days, config, 95.0)
    result = atmos.fraction_over_precip_thresh(
        config.cf_variables[0].da,
        per,
        thresh="1 mm/day",
        freq=config.freq.panda_freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    result = result * 100
    result.attrs["units"] = "%"
    if config.save_percentile:
        return result, per
    return result, None


def r99p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(
        config.cf_variables[0].in_base_da, dry_day_value=np.nan
    )
    per = _compute_percentile(base_wet_days, config, 99.0)
    result = atmos.days_over_precip_thresh(
        config.cf_variables[0].da,
        per,
        thresh="1 mm/day",
        freq=config.freq.panda_freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if config.save_percentile:
        return result, per
    if config.is_percent:
        result = _to_percent(result, config.freq)
    return result, None


def r99ptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(
        config.cf_variables[0].in_base_da, dry_day_value=np.nan
    )
    per = _compute_percentile(base_wet_days, config, 99.0)
    result = atmos.fraction_over_precip_thresh(
        config.cf_variables[0].da,
        per,
        thresh="1 mm/day",
        freq=config.freq.panda_freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    result = result * 100
    result.attrs["units"] = "%"
    if config.save_percentile:
        return result, per
    return result, None


def sd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = land.snow_depth(config.cf_variables[0].da, freq=config.freq.panda_freq)
    return result, None


def sd1(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = land.snow_cover_duration(
        config.cf_variables[0].da, thresh="1 cm", freq=config.freq.panda_freq
    )
    return result, None


def sd5cm(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = land.snow_cover_duration(
        config.cf_variables[0].da, thresh="5 cm", freq=config.freq.panda_freq
    )
    return result, None


def sd50cm(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = land.snow_cover_duration(
        config.cf_variables[0].da, thresh="50 cm", freq=config.freq.panda_freq
    )
    return result, None


def tg(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tg_mean(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def tn(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tn_mean(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def tx(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.tx_mean(config.cf_variables[0].da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "degC")
    return result, None


def dtr(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.daily_temperature_range(
        tasmax=config.cf_variables[0].da,
        tasmin=config.cf_variables[1].da,
        freq=config.freq.panda_freq,
    )
    result.attrs["units"] = "°C"
    return result, None


def etr(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.extreme_temperature_range(
        tasmax=config.cf_variables[0].da,
        tasmin=config.cf_variables[1].da,
        freq=config.freq.panda_freq,
    )
    result.attrs["units"] = "°C"
    return result, None


def vdtr(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    result = atmos.daily_temperature_range_variability(
        tasmax=config.cf_variables[0].da,
        tasmin=config.cf_variables[1].da,
        freq=config.freq.panda_freq,
    )
    result.attrs["units"] = "°C"
    return result, None


def cd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    tas_per = _compute_percentile_doy(
        config.cf_variables[0].in_base_da, config, 25
    ).squeeze(PERCENTILES_COORD, drop=True)
    precip_cfvar = config.cf_variables[1]
    precip_cfvar.in_base_da = _filter_in_wet_days(
        precip_cfvar.in_base_da, dry_day_value=np.NAN
    )
    precip_cfvar.da = _filter_in_wet_days(precip_cfvar.da, dry_day_value=0)
    pr_per = _compute_percentile_doy(precip_cfvar.in_base_da, config, 25).squeeze(
        PERCENTILES_COORD, drop=True
    )
    result = atmos.cold_and_dry_days(
        tas=config.cf_variables[0].da,
        tas_25=tas_per,
        pr=precip_cfvar.da,
        pr_25=pr_per,
        freq=config.freq.panda_freq,
    )
    if config.save_percentile:
        result.coords["tas_per"] = resample_doy(tas_per, result)
        result.coords["pr_per"] = resample_doy(pr_per, result)
    return result, None


def cw(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    tas_per = _compute_percentile_doy(
        config.cf_variables[0].in_base_da, config, 25
    ).squeeze(PERCENTILES_COORD, drop=True)
    precip_cfvar = config.cf_variables[1]
    precip_cfvar.in_base_da = _filter_in_wet_days(
        precip_cfvar.in_base_da, dry_day_value=np.nan
    )
    precip_cfvar.da = _filter_in_wet_days(precip_cfvar.da, dry_day_value=0)
    pr_per = _compute_percentile_doy(precip_cfvar.in_base_da, config, 75).squeeze(
        PERCENTILES_COORD, drop=True
    )
    result = atmos.cold_and_wet_days(
        tas=config.cf_variables[0].da,
        tas_25=tas_per,
        pr=precip_cfvar.da,
        pr_75=pr_per,
        freq=config.freq.panda_freq,
    )
    if config.save_percentile:
        result.coords["tas_per"] = resample_doy(tas_per, result)
        result.coords["pr_per"] = resample_doy(pr_per, result)
    return result, None


def wd(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    tas_per = _compute_percentile_doy(
        config.cf_variables[0].in_base_da, config, 75
    ).squeeze(PERCENTILES_COORD, drop=True)
    precip_cfvar = config.cf_variables[1]
    precip_cfvar.in_base_da = _filter_in_wet_days(
        precip_cfvar.in_base_da, dry_day_value=np.nan
    )
    precip_cfvar.da = _filter_in_wet_days(precip_cfvar.da, dry_day_value=0)
    pr_per = _compute_percentile_doy(precip_cfvar.in_base_da, config, 25).squeeze(
        PERCENTILES_COORD, drop=True
    )
    result = atmos.warm_and_dry_days(
        tas=config.cf_variables[0].da,
        tas_75=tas_per,
        pr=precip_cfvar.da,
        pr_25=pr_per,
        freq=config.freq.panda_freq,
    )
    if config.save_percentile:
        result.coords["tas_per"] = resample_doy(tas_per, result)
        result.coords["pr_per"] = resample_doy(pr_per, result)
    return result, None


def ww(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    tas_per = _compute_percentile_doy(
        config.cf_variables[0].in_base_da, config, 75
    ).squeeze(PERCENTILES_COORD, drop=True)
    precip_cfvar = config.cf_variables[1]
    precip_cfvar.in_base_da = _filter_in_wet_days(
        precip_cfvar.in_base_da, dry_day_value=np.nan
    )
    precip_cfvar.da = _filter_in_wet_days(precip_cfvar.da, dry_day_value=0)
    pr_per = _compute_percentile_doy(precip_cfvar.in_base_da, config, 75).squeeze(
        PERCENTILES_COORD, drop=True
    )
    result = atmos.warm_and_wet_days(
        tas=config.cf_variables[0].da,
        tas_75=tas_per,
        pr=precip_cfvar.da,
        pr_75=pr_per,
        freq=config.freq.panda_freq,
    )
    if config.save_percentile:
        result.coords["tas_per"] = resample_doy(tas_per, result)
        result.coords["pr_per"] = resample_doy(pr_per, result)
    return result, None


def _add_celsius_suffix(threshold: Optional[Union[str, float, int]]) -> Optional[str]:
    if threshold is not None:
        return f"{threshold} degC"
    return None


def _can_run_bootstrap(config: IndexConfig, percentile_period) -> bool:
    overlapping_years = np.unique(
        config.cf_variables[0].da.time.sel(time=percentile_period).dt.year
    )
    # No bootstrap if there is one single year overlapping
    # or no year overlapping
    # or all year overlapping
    run_bootstrap = (
        config.cf_variables[0].in_base_da is not config.cf_variables[0].da
        and len(overlapping_years) > 1
    )
    if run_bootstrap:
        in_base = config.cf_variables[0].in_base_da
        chunking = {d: "auto" for d in in_base.dims}
        chunking["time"] = -1
        config.cf_variables[0].in_base_da = in_base.chunk(chunking)
        config.cf_variables[0].da = config.cf_variables[0].da.chunk(chunking)
    if run_bootstrap and config.interpolation != QuantileInterpolation.MEDIAN_UNBIASED:
        raise InvalidIcclimArgumentError(
            "When bootstrapping, the interpolation must be MEDIAN_UNBIASED."
            f" Here it was {config.interpolation}."
        )
    return run_bootstrap


def _to_percent(da: DataArray, sampling_freq: Frequency) -> DataArray:
    if sampling_freq == Frequency.MONTH:
        da = da / da.time.dt.daysinmonth * 100
    elif sampling_freq == Frequency.YEAR:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[leap_years] = 366
        coef[~leap_years] = 365
        da = da / coef
    elif sampling_freq == Frequency.AMJJAS:
        da = da / 183
    elif sampling_freq == Frequency.ONDJFM:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[leap_years] = 183
        coef[~leap_years] = 182
        da = da / coef
    elif sampling_freq == Frequency.DJF:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[leap_years] = 91
        coef[~leap_years] = 90
        da = da / coef
    elif sampling_freq in [Frequency.MAM, Frequency.JJA]:
        da = da / 92
    elif sampling_freq == Frequency.SON:
        da = da / 91
    else:
        # TODO improve this for custom resampling
        warn(
            "For now, '%' unit can only be used with slice_mode being one of "
            "{MONTH, YEAR, AMJJAS, ONDJFM, DJF, MAM, JJA, SON}."
        )
        return da
    da.attrs["units"] = "1"
    return da


def _is_leap_year(da: DataArray) -> np.ndarray:
    time_index = da.indexes.get("time")
    if isinstance(time_index, xr.CFTimeIndex):
        cf_calendar: CfCalendar = CfCalendar.lookup(time_index.calendar)
        return cf_calendar.is_leap(da.time.dt.year)
    else:
        return da.time.dt.is_leap_year


def _add_bootstrap_meta(result: DataArray, per: DataArray) -> DataArray:
    result.attrs[IN_BASE_IDENTIFIER] = per.climatology_bounds
    return result


def _compute_percentile_doy(
    da: DataArray, config: IndexConfig, percentile: int
) -> DataArray:
    per = percentile_doy(
        da,
        config.window,
        percentile,
        alpha=config.interpolation.alpha,
        beta=config.interpolation.beta,
    )
    if config.callback is not None:
        config.callback(50)
    return per


def _compute_percentile(
    arr: DataArray, config: IndexConfig, percentiles: float
) -> DataArray:
    return xr.apply_ufunc(
        xclim.core.utils.calc_perc,
        arr,
        input_core_dims=[["time"]],
        output_core_dims=[[PERCENTILES_COORD]],
        kwargs=dict(
            percentiles=[percentiles],
            alpha=config.interpolation.alpha,
            beta=config.interpolation.beta,
        ),
        dask="parallelized",
        output_dtypes=[arr.dtype],
        dask_gufunc_kwargs=dict(output_sizes={PERCENTILES_COORD: len([percentiles])}),
    )


def _filter_in_wet_days(da: DataArray, dry_day_value):
    """
    Turns non wet days to NaN.
    """
    precip = convert_units_to(da, "mm/d")
    return precip.where(precip > 1, dry_day_value)
